using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using MinorShift.Emuera.Sub;
using MinorShift.Emuera.GameData;
using MinorShift.Emuera.GameView;
using MinorShift.Emuera.GameData.Expression;
using MinorShift.Emuera.GameData.Variable;
using MinorShift.Emuera.GameData.Function;
namespace MinorShift.Emuera.GameProc
{
	internal abstract class Argument
	{
		//ArgumentType type = (ArgumentType)((int)ArgumentType.__MASK__ & (int)func);
	}

	internal sealed class VoidArgument : Argument
	{
		public VoidArgument() { }
	}

	internal sealed class ExpressionArgument : Argument
	{
		public ExpressionArgument(IOperandTerm termSrc)
		{
			Term = termSrc;
		}
		readonly public IOperandTerm Term;
	}

    internal sealed class ExpressionArrayArgument : Argument
    {
        public ExpressionArrayArgument(List<IOperandTerm> termList)
        {
            TermList = new IOperandTerm[termList.Count];
            termList.CopyTo(TermList);
        }
        readonly public IOperandTerm[] TermList;
    }

	//internal sealed class StrArgument : Argument
	//{
	//    public StrArgument(string str)
	//    {
	//        Str = str;
	//    }
	//    readonly public string Str;
	//}

	//internal sealed class FormStrArgument : Argument
	//{
	//    public FormStrArgument(StringForm strform)
	//    {
	//        StrForm = strform;
	//    }
	//    readonly public StringForm StrForm;
	//}


	internal sealed class SpPrintVArgument : Argument
	{
		public SpPrintVArgument(List<IOperandTerm> list)
		{
			Terms = list;
		}
		readonly public List<IOperandTerm> Terms;
	}

	internal sealed class SpTimesArgument : Argument
	{
		public SpTimesArgument(VariableToken var, double d)
		{
			VariableDest = var;
			DoubleValue = d;
		}
		readonly public VariableToken VariableDest;
		readonly public double DoubleValue;
	}

	internal sealed class SpBarArgument : Argument
	{
		public SpBarArgument(IOperandTerm value, IOperandTerm max, IOperandTerm length)
		{
			Terms[0] = value;
			Terms[1] = max;
			Terms[2] = length;
		}
		readonly public IOperandTerm[] Terms = new IOperandTerm[3];
	}

	internal sealed class SpSetArgument : Argument
	{
		public SpSetArgument(VariableToken var, IOperandTerm termSrc)
		{
			VariableDest = var;
			Term = termSrc;
		}
		readonly public VariableToken VariableDest;
		readonly public IOperandTerm Term;
	}

	internal sealed class SpSetsArgument : Argument
	{
		public SpSetsArgument(VariableToken var, StringForm strform)
		{
			VariableDest = var;
			StrForm = strform;
		}
		readonly public VariableToken VariableDest;
		readonly public StringForm StrForm;
	}

	//internal sealed class SpSubstringArgument : Argument
	//{
	//    public SpSubstringArgument(IOperandTerm var, IOperandTerm start, IOperandTerm length)
	//    {
	//        StrExpression = var;
	//        Start = start;
	//        Length = length;
	//    }
	//    readonly public IOperandTerm StrExpression;
	//    readonly public IOperandTerm Start;
	//    readonly public IOperandTerm Length;
	//}

	internal sealed class SpSwapCharaArgument : Argument
	{
		public SpSwapCharaArgument(IOperandTerm x, IOperandTerm y)
		{
			X = x;
			Y = y;
		}
		readonly public IOperandTerm X;
		readonly public IOperandTerm Y;
	}

	internal sealed class SpSwapVarArgument : Argument
	{
		public SpSwapVarArgument(VariableToken v1, VariableToken v2)
		{
			var1 = v1;
			var2 = v2;
		}
		readonly public VariableToken var1;
		readonly public VariableToken var2;
	}

	internal sealed class SpVarsizeArgument : Argument
	{
		public SpVarsizeArgument(VariableIdentifier var)
		{
			VariableID = var;
		}
		readonly public VariableIdentifier VariableID;
	}

	internal sealed class SpSaveDataArgument : Argument
	{
		public SpSaveDataArgument(IOperandTerm target, IOperandTerm var)
		{
			Target = target;
			StrExpression = var;
		}
		readonly public IOperandTerm Target;
		readonly public IOperandTerm StrExpression;
	}

	internal sealed class SpTInputsArgument : Argument
	{
		public SpTInputsArgument(IOperandTerm time, IOperandTerm def, IOperandTerm disp)
		{
			Time = time;
			Def = def;
			Disp = disp;
		}
		readonly public IOperandTerm Time;
		readonly public IOperandTerm Def;
		readonly public IOperandTerm Disp;
	}

	internal enum SortOrder
	{
		UNDEF = 0,
		ASCENDING = 1,
		DESENDING = 2,
	}

	internal sealed class SpSortcharaArgument : Argument
	{
		public SpSortcharaArgument(VariableToken var, SortOrder order)
		{
			SortKey = var;
			SortOrder = order;
		}
		readonly public VariableToken SortKey;
		readonly public SortOrder SortOrder;
	}

	internal sealed class SpCallArgment : Argument
	{
		public SpCallArgment(string funcname, IOperandTerm[] args)
		{
			Str = funcname;
			Args = args;
		}
		readonly public string Str;
		readonly public IOperandTerm[] Args;
	}

	internal sealed class SpCallformArgment : Argument
	{
		public SpCallformArgment(StringForm formstr, IOperandTerm[] args)
		{
			StrForm = formstr;
			Args = args;
		}
		readonly public StringForm StrForm;
		readonly public IOperandTerm[] Args;
	}

	internal sealed class SpForNextArgment : Argument
	{
		public SpForNextArgment(VariableToken var, IOperandTerm start, IOperandTerm end, IOperandTerm step)
		{
			this.Cnt = var;
			this.Start = start;
			this.End = end;
			this.Step = step;
		}
		readonly public VariableToken Cnt;
		readonly public IOperandTerm Start;
		readonly public IOperandTerm End;
		readonly public IOperandTerm Step;
	}

	internal sealed class SpPowerArgument : Argument
	{
		public SpPowerArgument(VariableToken var, IOperandTerm x, IOperandTerm y)
		{
			VariableDest = var;
			X = x;
			Y = y;
		}
		readonly public VariableToken VariableDest;
		readonly public IOperandTerm X;
		readonly public IOperandTerm Y;
	}

	//internal sealed class SpStrFindArgument : Argument
	//{
	//    public SpStrFindArgument(IOperandTerm target, IOperandTerm word, IOperandTerm start)
	//    {
	//        TargetStr = target;
	//        WordStr = word;
	//        StartIndex = start;
	//    }
	//    readonly public IOperandTerm TargetStr;
	//    readonly public IOperandTerm WordStr;
	//    readonly public IOperandTerm StartIndex;
	//}

	//internal sealed class SpFindCharaArgument : Argument
	//{
	//    public SpFindCharaArgument(VariableToken var, IOperandTerm word, IOperandTerm start)
	//    {
	//        CharaVar = var;
	//        Word = word;
	//        StartIndex = start;
	//    }
	//    readonly public VariableToken CharaVar;
	//    readonly public IOperandTerm Word;
	//    readonly public IOperandTerm StartIndex;
	//}

	internal sealed class CaseArgument : Argument
	{
		public CaseArgument(CaseExpression[] args)
		{
			CaseExps = args;
		}
		readonly public CaseExpression[] CaseExps;
	}

    internal sealed class PrintDataArgument : Argument
    {
        public PrintDataArgument(VariableToken var)
        {
            Var = var;
        }
        readonly public VariableToken Var;
    }

	internal sealed class FunctionArgument : Argument
	{
		public FunctionArgument(IOperandTerm method)
		{
			MethodTerm = method;
		}
		readonly public IOperandTerm MethodTerm;
	}

    internal sealed class BitArgument : Argument
    {
        public BitArgument(VariableToken var, IOperandTerm termSrc)
        {
            VariableDest = var;
            Term = termSrc;
        }
        readonly public VariableToken VariableDest;
        readonly public IOperandTerm Term;
	}

	internal sealed class SpVarSetArgument : Argument
	{
		public SpVarSetArgument(VariableToken var, IOperandTerm termSrc, IOperandTerm start, IOperandTerm end)
		{
			VariableDest = var;
			Term = termSrc;
            Start = start;
            End = end;
		}
		readonly public VariableToken VariableDest;
		readonly public IOperandTerm Term;
        readonly public IOperandTerm Start;
        readonly public IOperandTerm End;
	}

    internal sealed class SpCVarSetArgument : Argument
    {
        public SpCVarSetArgument(VariableToken var, IOperandTerm indexTerm, IOperandTerm termSrc, IOperandTerm start, IOperandTerm end)
        {
            VariableDest = var;
            Index = indexTerm;
            Term = termSrc;
            Start = start;
            End = end;
        }
        readonly public VariableToken VariableDest;
        readonly public IOperandTerm Index;
        readonly public IOperandTerm Term;
        readonly public IOperandTerm Start;
        readonly public IOperandTerm End;
    }

	internal sealed class SpButtonArgument : Argument
	{
		public SpButtonArgument(IOperandTerm p1, IOperandTerm p2)
		{
			PrintStrTerm = p1;
			ButtonWord = p2;
		}
		readonly public IOperandTerm PrintStrTerm;
		readonly public IOperandTerm ButtonWord;
	}

	internal sealed class SpSetArrayArgument : Argument
	{
		public SpSetArrayArgument(VariableToken var, IOperandTerm[] termList)
		{
			VariableDest = var;
			TermList = termList;
		}
		readonly public VariableToken VariableDest;
		readonly public IOperandTerm[] TermList;
	}

	internal sealed class SpColorArgument : Argument
	{
		public SpColorArgument(IOperandTerm r,IOperandTerm g, IOperandTerm b)
		{
			R = r;
			G = g;
			B = b;
		}
		public SpColorArgument(IOperandTerm rgb)
		{
			RGB = rgb;
		}
		readonly public IOperandTerm R;
		readonly public IOperandTerm G;
		readonly public IOperandTerm B;
		readonly public IOperandTerm RGB;
	}

    internal sealed class SpSplitArgument : Argument
    {
        public SpSplitArgument(IOperandTerm s1, IOperandTerm s2, VariableIdentifier varId)
        {
            TargetStr = s1;
            Split = s2;
            Var = varId;
        }
        readonly public IOperandTerm TargetStr;
        readonly public IOperandTerm Split;
        readonly public VariableIdentifier Var;
    }

    internal sealed class SpGetIntArgument : Argument
    {
        public SpGetIntArgument(VariableToken var)
        {
            VarToken = var;
        }
        readonly public VariableToken VarToken;
    }

    internal sealed class SpArrayControlArgument : Argument
    {
        public SpArrayControlArgument(VariableToken var, IOperandTerm num1, IOperandTerm num2)
        {
            VarToken = var;
            Num1 = num1;
            Num2 = num2;
        }
        readonly public VariableToken VarToken;
        readonly public IOperandTerm Num1;
        readonly public IOperandTerm Num2;
    }

    internal sealed class SpArrayShiftArgument : Argument
    {
        public SpArrayShiftArgument(VariableToken var, IOperandTerm num1, IOperandTerm num2, IOperandTerm num3, IOperandTerm num4)
        {
            VarToken = var;
            Num1 = num1;
            Num2 = num2;
            Num3 = num3;
            Num4 = num4;
        }
        readonly public VariableToken VarToken;
        readonly public IOperandTerm Num1;
        readonly public IOperandTerm Num2;
        readonly public IOperandTerm Num3;
        readonly public IOperandTerm Num4;
    }

}
