using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using Microsoft.VisualBasic;
using MinorShift.Emuera.Sub;
using MinorShift.Emuera.GameData;
using MinorShift.Emuera.GameData.Expression;
using MinorShift.Emuera.GameData.Variable;
using MinorShift._Library;
using MinorShift.Emuera.GameView;
using MinorShift.Emuera.GameData.Function;

namespace MinorShift.Emuera.GameProc
{
	internal sealed partial class Process
	{
		private void runScriptProc()
		{
            while(true)
            {
                bool sequential = state.Sequential;
                state.ShfitNextLine();
                LogicalLine line = state.CurrentLine;
                vEvaluator.Scope = state.Scope;
                if (line == null)
                    throw new ExeEE("Emuera.exeは次に実行する行を見失いました");
                else if (line.IsError)
                    throw new CodeEE(line.ErrMes);
                else if (line is InstructionLine)
                {//1753 InstructionLineを先に持ってきてみる。わずかに速くなった気がしないでもない
                    InstructionLine func = (InstructionLine)line;
                    InstructionLine jumpTo = null;

					if (!Program.DebugMode && func.Function.IsDebug())
					{//非DebugモードでのDebug系命令。何もしない。（SIF文のためにコメント行扱いにはできない）
						continue;
					}
                    if (func.Argument == null)
                        LogicalLineParser.SetArgumentTo(func, null);
                    if (func.IsError)
                        throw new CodeEE(func.ErrMes);
					FunctionIdentifier iFunc = func.Function;
					FunctionCode iFuncCode = func.FunctionCode;
					if (!iFunc.IsFlowContorol())
                    {
                        #region fastPathNoramlFunction
                        if (iFunc.IsMethod())
                        {
                            FunctionArgument arg = (FunctionArgument)func.Argument;
                            Type type = arg.MethodTerm.GetOperandType();
                            if (arg.MethodTerm.GetOperandType() == typeof(Int64))
                                vEvaluator.RESULT = eEvaluator.GetInteger(arg.MethodTerm);
                            else if (arg.MethodTerm.GetOperandType() == typeof(string))
                                vEvaluator.RESULTS = eEvaluator.GetString(arg.MethodTerm);
                            else
                                throw new ExeEE(func.FunctionCode.ToString() + "命令の型が不明");
                        }
                        else if (iFuncCode == FunctionCode.SET)
                        {
                            SpSetArgument spsetarg = (SpSetArgument)func.Argument;
                            FixedVariablePointer p = eEvaluator.GetFixedVariable(spsetarg.VariableDest);
                            vEvaluator.SetValue(p, eEvaluator.GetInteger(spsetarg.Term));
                        }
                        else if (iFuncCode == FunctionCode.SETS)
                        {
                            SpSetsArgument spsetsarg = (SpSetsArgument)func.Argument;
                            FixedVariablePointer p = eEvaluator.GetFixedVariable(spsetsarg.VariableDest);
                            vEvaluator.SetValue(p, spsetsarg.StrForm.GetString(eEvaluator));
						}
                        #endregion
                        else
                            doNormalFunction(func);
                    }
                    else
                    {
                        #region fastPathFlowFunction
                        if (iFuncCode == FunctionCode.SIF || iFuncCode == FunctionCode.IF || iFuncCode == FunctionCode.SELECTCASE
                            || iFuncCode == FunctionCode.ELSEIF || iFuncCode == FunctionCode.ELSE || iFuncCode == FunctionCode.CASE
                            || iFuncCode == FunctionCode.CASEELSE
                            || iFuncCode == FunctionCode.ENDIF || iFuncCode == FunctionCode.ENDSELECT || iFuncCode == FunctionCode.ENDCATCH
                            || iFuncCode == FunctionCode.ENDFUNC || iFuncCode == FunctionCode.DO)
                        {
                            ExpressionArgument expArg = null;
                            if (iFuncCode == FunctionCode.ELSE || iFuncCode == FunctionCode.ELSEIF
                                || iFuncCode == FunctionCode.CASE || iFuncCode == FunctionCode.CASEELSE)
                            {
                                if (func.JumpTo == null)
                                    throw new ExeEE(func.FunctionCode.ToString() + "のジャンプ先が設定されていない");
                                state.JumpTo(func.JumpTo);
                            }
                            else if (iFuncCode == FunctionCode.SIF)
                            {
                                expArg = (ExpressionArgument)func.Argument;
                                if (eEvaluator.GetInteger(expArg.Term) == 0)//評価式が真ならそのまま流れ落ちる
                                    state.ShfitNextLine();//偽なら一行とばす。順に来たときと同じ扱いにする
                            }
                            else if (iFuncCode == FunctionCode.IF)
                            {
                                LogicalLine ifJumpto = null;
                                if (func.IfCaseList == null)
                                    throw new ExeEE("IFのIF-ELSEIFリストが適正に作成されていない");
                                if (func.JumpTo == null)
                                    throw new ExeEE("IFに対応するENDIFが設定されていない");

                                for (int i = 0; i < func.IfCaseList.Count; i++)
                                {
                                    if (func.IfCaseList[i].IsError)
                                        continue;
									if (func.IfCaseList[i].FunctionCode == FunctionCode.ELSE)
                                    {
                                        ifJumpto = func.IfCaseList[i];
                                        break;
                                    }

                                    expArg = (ExpressionArgument)(func.IfCaseList[i].Argument);
                                    if (expArg == null)
                                        throw new ExeEE("IFチェック中。引数が解析されていない。", func.IfCaseList[i].Position);

                                    //1730 ELSEIFが出したエラーがIFのエラーとして検出されていた
                                    state.RunningLine = func.IfCaseList[i];
                                    if (eEvaluator.GetInteger(expArg.Term) != 0)//式が真
                                    {
                                        ifJumpto = func.IfCaseList[i];
                                        break;
                                    }
                                }
                                if (ifJumpto == null)
                                    ifJumpto = func.JumpTo;//ENDIF
                                if (ifJumpto != func)//自分自身がジャンプ先ならそのまま
                                    state.JumpTo(ifJumpto);
                                state.RunningLine = null;
                            }
                            else if (iFuncCode == FunctionCode.SELECTCASE)
                            {
                                LogicalLine caseJumpto = null;
                                SingleTerm term = eEvaluator.GetValue(((ExpressionArgument)func.Argument).Term);
                                if (func.IfCaseList == null)
                                    throw new ExeEE("SELECTCASEのCASEリストが適正に作成されていない");
                                if (func.JumpTo == null)
                                    throw new ExeEE("SELECTCASEに対応するENDSELECTが設定されていない");
                                for (int i = 0; i < func.IfCaseList.Count; i++)
                                {
                                    if (func.IfCaseList[i].IsError)
                                        continue;
									if (func.IfCaseList[i].FunctionCode == FunctionCode.CASEELSE)
                                    {
                                        caseJumpto = func.IfCaseList[i];
                                        break;
                                    }
                                    CaseArgument caseArg = (CaseArgument)(func.IfCaseList[i].Argument);
                                    if (caseArg == null)
                                        throw new ExeEE("CASEチェック中。引数が解析されていない。", func.IfCaseList[i].Position);

                                    //1730 CASEが出したエラーがSELECTCASEのエラーとして検出されていた
                                    state.RunningLine = func.IfCaseList[i];
									if (term.GetOperandType() == typeof(Int64))
									{
										Int64 Is = term.Int;
										foreach (CaseExpression caseExp in caseArg.CaseExps)
										{
											if (eEvaluator.GetCase(caseExp, Is))
											{
												caseJumpto = func.IfCaseList[i];
												goto casefound;
											}
										}
									}
									else
									{
										string Is = term.Str;
										foreach (CaseExpression caseExp in caseArg.CaseExps)
										{
											if (eEvaluator.GetCase(caseExp, Is))
											{
												caseJumpto = func.IfCaseList[i];
												goto casefound;
											}
										}
									}

                                }
                            casefound:
                                if (caseJumpto == null)
                                    caseJumpto = func.JumpTo;//ENDSELECT
                                state.JumpTo(caseJumpto);
                                state.RunningLine = null;
                            }
                        }
                        else if (iFuncCode == FunctionCode.RETURN|| iFuncCode == FunctionCode.RETURNFORM || iFuncCode == FunctionCode.CONTINUE
                            || iFuncCode == FunctionCode.RESTART|| iFuncCode == FunctionCode.BREAK|| iFuncCode == FunctionCode.REND
                            || iFuncCode == FunctionCode.NEXT || iFuncCode == FunctionCode.WEND || iFuncCode == FunctionCode.LOOP)
                        {
                            if (iFuncCode == FunctionCode.REND || iFuncCode == FunctionCode.NEXT
								|| (iFuncCode == FunctionCode.CONTINUE && ((((InstructionLine)func.JumpTo).FunctionCode == FunctionCode.REPEAT) || ((InstructionLine)func.JumpTo).FunctionCode == FunctionCode.FOR)))
                            {
                                jumpTo = (InstructionLine)func.JumpTo;
                                //ループ変数が不明(REPEAT、FORを経由せずにループしようとした場合は無視してループを抜ける(eramakerがこういう仕様だったりする))
                                if (jumpTo.LoopCounter == null)
                                    state.JumpTo(jumpTo.JumpTo);
                                else
                                {
                                    unchecked
                                    {
                                        vEvaluator.SetValue(jumpTo.LoopCounter, vEvaluator.GetInteger(jumpTo.LoopCounter) + jumpTo.LoopStep);
                                    }
                                    if (((jumpTo.LoopStep > 0) && (jumpTo.LoopEnd > vEvaluator.GetInteger(jumpTo.LoopCounter)))
                                        || ((jumpTo.LoopStep < 0) && (jumpTo.LoopEnd < vEvaluator.GetInteger(jumpTo.LoopCounter))))//まだ回数が残っているなら、
                                        state.JumpTo(func.JumpTo);
									else if (func.FunctionCode == FunctionCode.CONTINUE)
                                        state.JumpTo(jumpTo.JumpTo);
                                }
                            }
							else if (iFuncCode == FunctionCode.WEND || (iFuncCode == FunctionCode.CONTINUE && ((InstructionLine)func.JumpTo).FunctionCode == FunctionCode.WHILE))
                            {
                                jumpTo = (InstructionLine)func.JumpTo;
                                if (eEvaluator.GetInteger(((ExpressionArgument)jumpTo.Argument).Term) != 0)
                                    state.JumpTo(func.JumpTo);
								else if (func.FunctionCode == FunctionCode.CONTINUE)
                                    state.JumpTo(jumpTo.JumpTo);
                            }
							else if (iFuncCode == FunctionCode.LOOP || (iFuncCode == FunctionCode.CONTINUE && ((InstructionLine)func.JumpTo).FunctionCode == FunctionCode.DO))
                            {
                                InstructionLine tFunc = func;
                                if (iFuncCode == FunctionCode.CONTINUE)
                                    tFunc = (InstructionLine)((InstructionLine)func.JumpTo).JumpTo;
                                if (tFunc.Argument == null)
                                    LogicalLineParser.SetArgumentTo(tFunc, null);
                                ExpressionArgument expArg = (ExpressionArgument)tFunc.Argument;
                                if (eEvaluator.GetInteger(expArg.Term) != 0)//式が真
                                    state.JumpTo(tFunc.JumpTo);
                                else if (iFuncCode == FunctionCode.CONTINUE)
                                    state.JumpTo((InstructionLine)((InstructionLine)func.JumpTo).JumpTo);
                            }
                            else if (iFuncCode == FunctionCode.BREAK)
                            {
                                ////BREAKのJUMP先はRENDまたはNEXT。そのジャンプ先であるREPEATかFORをiLineに代入。
                                //1.723 仕様変更。BREAKのJUMP先にはREPEAT、FOR、WHILEを記憶する。そのJUMP先が本当のJUMP先。
                                jumpTo = (InstructionLine)func.JumpTo;
                                InstructionLine iLine = (InstructionLine)jumpTo.JumpTo;
                                //WHILEとDOはカウンタがないので、即ジャンプ
								if (jumpTo.FunctionCode != FunctionCode.WHILE && jumpTo.FunctionCode != FunctionCode.DO)
                                {
                                    unchecked
                                    {//本家ではBREAK時にCOUNTが回る
                                        vEvaluator.SetValue(jumpTo.LoopCounter, vEvaluator.GetInteger(jumpTo.LoopCounter) + jumpTo.LoopStep);
                                    }
                                }
                                state.JumpTo(iLine);
                            }
                            else if (iFuncCode == FunctionCode.RESTART)
                                state.JumpTo(func.JumpTo);
							else if (iFuncCode == FunctionCode.RETURN)
							{
								int termnum = 0;
								foreach (IOperandTerm term in ((ExpressionArrayArgument)func.Argument).TermList)
								{
									vEvaluator.SetResultX(eEvaluator.GetInteger(term), termnum);
									termnum++;
								}
								state.Return(vEvaluator.RESULT);
								if (state.GetCurrentReturnAddress is NullLine)
									return;
							}
							else if (iFuncCode == FunctionCode.RETURNFORM)
							{
								int termnum = 0;
								foreach (IOperandTerm term in ((ExpressionArrayArgument)func.Argument).TermList)
								{
									string arg = eEvaluator.GetString(term);
									StringStream aSt = new StringStream(arg);
									vEvaluator.SetResultX(eEvaluator.GetInteger(ExpressionParser.ReduceIntegerTerm(aSt, null)), termnum);
									termnum++;
								}
								state.Return(vEvaluator.RESULT);
								if (state.GetCurrentReturnAddress is NullLine)
									return;
							}
                        }
                        #endregion
                        else if (!doFlowControlFunction(func, sequential))
                            return;
                    }
                    //lineCount++;
                    //WinmmTimerから時間を取得するのはそれ自体結構なコストがかかるので10000行に一回くらいで。
                    if ((lineCount % 10000 == 0 || (lineCount - 1) % 10000 == 0) && Config.Instance.InfiniteLoopAlertTime > 0)
                        checkInfiniteLoop();
                    if (console.IsRunning && !state.ScriptEnd)
                        continue;
                    return;
                }
				else if (line is NullLine)
				{//ファイル終端
					if (!state.IsFunctionMethod)
						vEvaluator.RESULT = 0;
					state.Return(0);
					if (!(state.GetCurrentReturnAddress is NullLine))
						continue;
					return;
				}
				else if (line is FunctionLabelLine)
				{
					if (sequential)
					{//流れ落ちてきた（関数終端）
						if (!state.IsFunctionMethod)
							vEvaluator.RESULT = 0;
						state.Return(0);
						if (!(state.GetCurrentReturnAddress is NullLine))
							continue;
						return;
					}
					//1750 飛んできた直後にShiftNextが入るのでここが実行されることは無いはず
					else//CALLやJUMPで飛んできた
						return;
				}
				else if (line is GotoLabelLine)
					continue;//＄ラベル。何もすることはない。
                else if (line is InvalidLine)
                {
                    if (string.IsNullOrEmpty(line.ErrMes))
                        throw new CodeEE("読込に失敗した行が実行されました。エラーの詳細は読込時の警告を参照してください。");
                    else
                        throw new CodeEE(line.ErrMes);
                }
                else
                    throw new ExeEE("定義されていない種類の行です");
            }
		}

		public void DoDebugNormalFunction(InstructionLine line, bool munchkin)
		{
			if (line.Function.IsMethod())
            {
                FunctionArgument arg = (FunctionArgument)line.Argument;
                Type type = arg.MethodTerm.GetOperandType();
                if (arg.MethodTerm.GetOperandType() == typeof(Int64))
                    vEvaluator.RESULT = eEvaluator.GetInteger(arg.MethodTerm);
                else if (arg.MethodTerm.GetOperandType() == typeof(string))
                    vEvaluator.RESULTS = eEvaluator.GetString(arg.MethodTerm);
                else
                    throw new ExeEE(line.FunctionCode.ToString() + "命令の型が不明");
            }
            else if (line.FunctionCode == FunctionCode.SET)
            {
                SpSetArgument spsetarg = (SpSetArgument)line.Argument;
                FixedVariablePointer p = eEvaluator.GetFixedVariable(spsetarg.VariableDest);
                vEvaluator.SetValue(p, eEvaluator.GetInteger(spsetarg.Term));
            }
			else if (line.FunctionCode == FunctionCode.SETS)
            {
                SpSetsArgument spsetsarg = (SpSetsArgument)line.Argument;
                FixedVariablePointer p = eEvaluator.GetFixedVariable(spsetsarg.VariableDest);
                vEvaluator.SetValue(p, spsetsarg.StrForm.GetString(eEvaluator));
            }
            else
                doNormalFunction(line);
			if (munchkin)
	            vEvaluator.IamaMunchkin();
        }

		#region normal
		void doNormalFunction(InstructionLine line)
		{
            //if ((line.Function & FunctionCode.__ARG_METHOD__) == FunctionCode.__ARG_METHOD__)
            //{
            //    FunctionArgument arg = (FunctionArgument)line.Argument;
            //    Type type = arg.MethodTerm.GetOperandType();
            //    if (arg.MethodTerm.GetOperandType() == typeof(Int64))
            //        vEvaluator.RESULT = eEvaluator.GetInteger(arg.MethodTerm);
            //    else if (arg.MethodTerm.GetOperandType() == typeof(string))
            //        vEvaluator.RESULTS = eEvaluator.GetString(arg.MethodTerm);
            //    else
            //        throw new ExeEE(line.Function.ToString() + "命令の型が不明");
            //    return;
            //}
			string str = null;
			IOperandTerm term = null;
			switch (line.FunctionCode)
			{
                //case FunctionCode.SET:
                //    {
                //        SpSetArgument spsetarg = (SpSetArgument)line.Argument;
                //        FixedVariablePointer p = eEvaluator.GetFixedVariable(spsetarg.VariableDest);
                //        vEvaluator.SetValue(p, eEvaluator.GetInteger(spsetarg.Term));
                //    }
                //    break;
                //case FunctionCode.SETS:
                //    {
                //        SpSetsArgument spsetsarg = (SpSetsArgument)line.Argument;
                //        FixedVariablePointer p = eEvaluator.GetFixedVariable(spsetsarg.VariableDest);
                //        vEvaluator.SetValue(p, spsetsarg.StrForm.GetString(eEvaluator));
                //    }
                //    break;
				case FunctionCode.PRINT://文字を表示する
                case FunctionCode.PRINTD:
                case FunctionCode.PRINTL://改行
                case FunctionCode.PRINTDL:
				case FunctionCode.PRINTW://入力待ち(実質改行)
                case FunctionCode.PRINTDW:
				case FunctionCode.PRINTSINGLE://強制一行
                case FunctionCode.PRINTSINGLED:
				case FunctionCode.PRINTS://文字列変数の内容
                case FunctionCode.PRINTSD:
                case FunctionCode.PRINTSL:
                case FunctionCode.PRINTSDL:
				case FunctionCode.PRINTSW:
                case FunctionCode.PRINTSDW:
				case FunctionCode.PRINTSINGLES:
                case FunctionCode.PRINTSINGLESD:
				case FunctionCode.PRINTFORM://{数式}、%文字列変数%などの書式が使える。
                case FunctionCode.PRINTFORMD:
				case FunctionCode.PRINTFORML:
                case FunctionCode.PRINTFORMDL:
				case FunctionCode.PRINTFORMW:
                case FunctionCode.PRINTFORMDW:
				case FunctionCode.PRINTSINGLEFORM:
                case FunctionCode.PRINTSINGLEFORMD:
                case FunctionCode.PRINTK:
                case FunctionCode.PRINTKL:
                case FunctionCode.PRINTKW:
                case FunctionCode.PRINTSINGLEK:
                case FunctionCode.PRINTSK:
                case FunctionCode.PRINTSKL:
                case FunctionCode.PRINTSKW:
                case FunctionCode.PRINTSINGLESK:
                case FunctionCode.PRINTFORMK:
                case FunctionCode.PRINTFORMKL:
                case FunctionCode.PRINTFORMKW:
                case FunctionCode.PRINTSINGLEFORMK:
                    {
                        if (SkipPrint)
                            break;
						console.UseUserStyle = !line.Function.IsPrintDFunction();
						term = ((ExpressionArgument)line.Argument).Term;
						str = eEvaluator.GetString(term);
						if (line.Function.IsPrintKFunction() && (forceHiragana | forceKatakana | halftoFull))
                            str = convertStringType(str);
						outputToConsole(str, line.Function);
					}
					break;

				case FunctionCode.PRINTV://変数の内容
                case FunctionCode.PRINTVD:
				case FunctionCode.PRINTVL:
                case FunctionCode.PRINTVDL:
				case FunctionCode.PRINTVW:
                case FunctionCode.PRINTVDW:
				case FunctionCode.PRINTSINGLEV:
                case FunctionCode.PRINTSINGLEVD:
                case FunctionCode.PRINTVK:
                case FunctionCode.PRINTVKL:
                case FunctionCode.PRINTVKW:
                case FunctionCode.PRINTSINGLEVK:
					{
                        if (SkipPrint)
                            break;
						console.UseUserStyle = !line.Function.IsPrintDFunction();
                        StringBuilder builder = new StringBuilder();
						foreach (IOperandTerm termV in ((SpPrintVArgument)line.Argument).Terms)
						{
							if (termV.GetOperandType() == typeof(Int64))
								builder.Append(eEvaluator.GetInteger(termV).ToString());
							else
								builder.Append(((SingleTerm)termV).Str);
						}
						str = builder.ToString();
						if (line.Function.IsPrintKFunction() && (forceHiragana | forceKatakana | halftoFull))
                            str = convertStringType(str);
						outputToConsole(str, line.Function);
                    }
					break;
				case FunctionCode.PRINTFORMS://文字列変数の内容を変換して表示をします。
                case FunctionCode.PRINTFORMSD:
				case FunctionCode.PRINTFORMSL:
                case FunctionCode.PRINTFORMSDL:
				case FunctionCode.PRINTFORMSW:
                case FunctionCode.PRINTFORMSDW:
				case FunctionCode.PRINTSINGLEFORMS:
                case FunctionCode.PRINTSINGLEFORMSD:
                case FunctionCode.PRINTFORMSK:
                case FunctionCode.PRINTFORMSKL:
                case FunctionCode.PRINTFORMSKW:
                case FunctionCode.PRINTSINGLEFORMSK:
					{
                        if (SkipPrint)
                            break;
						console.UseUserStyle = !line.Function.IsPrintDFunction();
                        term = ((ExpressionArgument)line.Argument).Term;
						str = eEvaluator.GetString(term).Replace("\\", "\\\\");
						StringForm strForm = new StringForm(str);
						str = strForm.GetString(eEvaluator);
						if (line.Function.IsPrintKFunction() && (forceHiragana | forceKatakana | halftoFull))
                            str = convertStringType(str);
						outputToConsole(str, line.Function);
                    }
					break;
				case FunctionCode.PRINTC:
                case FunctionCode.PRINTCD:
				case FunctionCode.PRINTLC:
                case FunctionCode.PRINTLCD:
				case FunctionCode.PRINTFORMC:
                case FunctionCode.PRINTFORMCD:
				case FunctionCode.PRINTFORMLC:
                case FunctionCode.PRINTFORMLCD:
                case FunctionCode.PRINTCK:
                case FunctionCode.PRINTLCK:
                case FunctionCode.PRINTFORMCK:
                case FunctionCode.PRINTFORMLCK:
					{
                        if (SkipPrint)
                            break;
						console.UseUserStyle = !line.Function.IsPrintDFunction();
                        term = ((ExpressionArgument)line.Argument).Term;
						str = eEvaluator.GetString(term);
						if (line.Function.IsPrintKFunction() && (forceHiragana | forceKatakana | halftoFull))
                            str = convertStringType(str);
                        if ((line.FunctionCode == FunctionCode.PRINTFORMC)
							|| (line.FunctionCode == FunctionCode.PRINTC)
							|| (line.FunctionCode == FunctionCode.PRINTCD)
							|| (line.FunctionCode == FunctionCode.PRINTFORMCD))
							console.PrintC(str, true);
						else
							console.PrintC(str, false);
                        console.UseUserStyle = true;
                    }
					break;
				case FunctionCode.PRINTBUTTON://変数の内容
					{
                        if (SkipPrint)
                            break;
                        SpButtonArgument bArg = (SpButtonArgument)line.Argument;
						str = eEvaluator.GetString(bArg.PrintStrTerm);
                        //ボタン処理に絡んで表示がおかしくなるため、PRINTBUTTONでの改行コードはオミット
                        str = str.Replace("\n", "");
                        if (bArg.ButtonWord.GetOperandType() == typeof(long))
							console.PrintButton(str,  eEvaluator.GetInteger(bArg.ButtonWord));
						else
							console.PrintButton(str, eEvaluator.GetString(bArg.ButtonWord));
					}
					break;
                case FunctionCode.PRINTBUTTONC://変数の内容
                case FunctionCode.PRINTBUTTONLC:
                    {
                        if (SkipPrint)
                            break;
                        SpButtonArgument bArg = (SpButtonArgument)line.Argument;
                        str = eEvaluator.GetString(bArg.PrintStrTerm);
                        //ボタン処理に絡んで表示がおかしくなるため、PRINTBUTTONでの改行コードはオミット
                        str = str.Replace("\n", "");
						bool isRight = (line.FunctionCode == FunctionCode.PRINTBUTTONC) ? true : false;
                        if (bArg.ButtonWord.GetOperandType() == typeof(long))
                            console.PrintButtonC(str, eEvaluator.GetInteger(bArg.ButtonWord), isRight);
                        else
                            console.PrintButtonC(str, eEvaluator.GetString(bArg.ButtonWord), isRight);
                    }
                    break;
                case FunctionCode.PRINTPLAIN:
                case FunctionCode.PRINTPLAINFORM:
                    {
                        if (SkipPrint)
                            break;
                        term = ((ExpressionArgument)line.Argument).Term;
                        console.PrintPlain(eEvaluator.GetString(term));
                    }
                    break;
                case FunctionCode.PRINTDATA:
                case FunctionCode.PRINTDATAD:
                case FunctionCode.PRINTDATAL:
                case FunctionCode.PRINTDATADL:
                case FunctionCode.PRINTDATAW:
                case FunctionCode.PRINTDATADW:
                case FunctionCode.PRINTDATAK:
                case FunctionCode.PRINTDATAKL:
                case FunctionCode.PRINTDATAKW:
                    {
                        if (SkipPrint)
                            break;
						console.UseUserStyle = !line.Function.IsPrintDFunction();
                        int count = line.dataList.Count;
                        int choice = (int)vEvaluator.GetNextRand(count);
                        VariableToken iTerm = ((PrintDataArgument)line.Argument).Var;
                        if (iTerm != null)
                            vEvaluator.SetValue(eEvaluator.GetFixedVariable(iTerm), choice);
                        List<InstructionLine> iList = line.dataList[choice];
                        int i = 0;
                        foreach (InstructionLine selectedLine in iList)
                        {
							state.RunningLine = selectedLine;
							if (selectedLine.Argument == null)
								LogicalLineParser.SetArgumentTo(selectedLine, null);
                            term = ((ExpressionArgument)selectedLine.Argument).Term;
                            str = eEvaluator.GetString(term);
							if (line.Function.IsPrintKFunction() && (forceHiragana | forceKatakana | halftoFull))
                                str = convertStringType(str);
                            console.Print(str);
                            if (++i < (int)iList.Count)
                                console.NewLine();
                        }
						if (line.Function.IsNewLine() || line.Function.IsWaitInput())
                            console.NewLine();
						if (line.Function.IsWaitInput())
                            console.ReadAnyKey();
                        console.UseUserStyle = true;
                        //ジャンプするが、流れが連続であることを保証。
                        state.JumpTo(line.JumpTo);
						state.RunningLine = null;
                    }
                    break;
                case FunctionCode.ENDDATA:
                    //何もしない
                    break;
				case FunctionCode.CLEARLINE:
					{
                        if (SkipPrint)
                            break;
                        ExpressionArgument intExpArg = (ExpressionArgument)line.Argument;
						Int32 delNum = (Int32)eEvaluator.GetInteger(intExpArg.Term);
						console.deleteLine(delNum);
						console.RefreshStrings(false);
					}
					break;
				case FunctionCode.REUSELASTLINE:
					{
                        if (SkipPrint)
                            break;
                        term = ((ExpressionArgument)line.Argument).Term;
						str = eEvaluator.GetString(term);
						console.PrintTemporaryLine(str);
					}
					break;
				case FunctionCode.WAIT://改行待ち。
                    if (SkipPrint)
                        break;
                    console.ReadAnyKey();
					break;
				case FunctionCode.TWAIT:
					{
                        if (SkipPrint)
                            break;
                        SpSwapCharaArgument arg = (SpSwapCharaArgument)line.Argument;
						Int64 time = eEvaluator.GetInteger(arg.X);
						Int64 flag = eEvaluator.GetInteger(arg.Y);
						console.waitInputWithTimer(time, flag);
						break;
					}
				case FunctionCode.INPUT://整数入力。入力はRESULTへ。
                    if (SkipPrint)
                    {
                        if (userDefinedSkip)
                        {
                            console.PrintError("表示スキップ中にINPUTに遭遇しました");
                            console.PrintError("INPUTに必要な処理をNOSKIP～ENDNOSKIPで囲むか、SKIPDISP 0～SKIPDISP 1で囲ってください");
                            throw new CodeEE("無限ループに入る可能性が高いため実行を終了します");
                        }
                        break;
                    }
                    console.ReadInteger();
					break;
                case FunctionCode.ONEINPUT:
                    if (SkipPrint)
                    {
                        if (userDefinedSkip)
                        {
                            console.PrintError("表示スキップ中にINPUTに遭遇しました");
                            console.PrintError("ONEINPUTに必要な処理をNOSKIP～ENDNOSKIPで囲むか、SKIPDISP 0～SKIPDISP 1で囲ってください");
                            throw new CodeEE("無限ループに入る可能性が高いため実行を終了します");
                        }
                        break;
                    }
                    console.ReadOneInteger();
                    break;
				case FunctionCode.INPUTS://文字列入力。入力はRESULTSへ。
                    if (SkipPrint)
                    {
                        if (userDefinedSkip)
                        {
                            console.PrintError("表示スキップ中にINPUTSに遭遇しました");
                            console.PrintError("INPUTSに必要な処理をNOSKIP～ENDNOSKIPで囲むか、SKIPDISP 0～SKIPDISP 1で囲ってください");
                            throw new CodeEE("無限ループに入る可能性が高いため実行を終了します");
                        }
                        break;
                    }
                    console.ReadString();
					break;
                case FunctionCode.ONEINPUTS:
                    if (SkipPrint)
                    {
                        if (userDefinedSkip)
                        {
                            console.PrintError("表示スキップ中にINPUTに遭遇しました");
                            console.PrintError("ONEINPUTSに必要な処理をNOSKIP～ENDNOSKIPで囲むか、SKIPDISP 0～SKIPDISP 1で囲ってください");
                            throw new CodeEE("無限ループに入る可能性が高いため実行を終了します");
                        }
                        break;
                    }
                    console.ReadOneString();
                    break;
                case FunctionCode.TINPUT:
					{
                        if (SkipPrint)
                        {
                            if (userDefinedSkip)
                            {
                                console.PrintError("表示スキップ中にTINPUTに遭遇しました");
                                console.PrintError("TINPUTに必要な処理をNOSKIP～ENDNOSKIPで囲むか、SKIPDISP 0～SKIPDISP 1で囲ってください");
                                console.PrintError("デフォルト値をRESULTに代入し処理を続行します");
                                vEvaluator.RESULT = eEvaluator.GetInteger(((SpBarArgument)line.Argument).Terms[1]);
                            }
                            break;
                        }
                        SpBarArgument arg = (SpBarArgument)line.Argument;
						long x = eEvaluator.GetInteger(arg.Terms[0]);
						long y = eEvaluator.GetInteger(arg.Terms[1]);
						long z = eEvaluator.GetInteger(arg.Terms[2]);
						console.ReadIntegerWithTimer(x, y, z);
                    }
					break;
				case FunctionCode.TINPUTS:
					{
                        if (SkipPrint)
                        {
                            if (userDefinedSkip)
                            {
                                console.PrintError("表示スキップ中にTINPUTSに遭遇しました");
                                console.PrintError("TINPUTSに必要な処理をNOSKIP～ENDNOSKIPで囲むか、SKIPDISP 0～SKIPDISP 1で囲ってください");
                                console.PrintError("デフォルト値をRESULTSに代入し処理を続行します");
                                vEvaluator.RESULTS = eEvaluator.GetString(((SpTInputsArgument)line.Argument).Def);
                            }
                            break;
                        }
                        SpTInputsArgument arg = (SpTInputsArgument)line.Argument;
						Int64 x = eEvaluator.GetInteger(arg.Time);
						string strs = eEvaluator.GetString(arg.Def);
						Int64 z = eEvaluator.GetInteger(arg.Disp);
						console.ReadStringWithTimer(x, strs, z);
                    }
					break;
				case FunctionCode.DRAWLINE://画面の左端から右端まで----と線を引く。
                    if (SkipPrint)
                        break;
                    console.PrintBar();
					console.NewLine();
					break;
				case FunctionCode.CUSTOMDRAWLINE:
					{
                        if (SkipPrint)
                            break;
                        term = ((ExpressionArgument)line.Argument).Term;
						str = eEvaluator.GetString(term);
						console.setStBar(str);
						console.PrintBar();
						console.NewLine();
						console.setStBar(Config.Instance.DrawLineString);
					}
					break;
				case FunctionCode.BAR://[*****....]のようなグラフを書く。BAR (変数) , (最大値), (長さ)
				case FunctionCode.BARL://改行付き。
					{
                        if (SkipPrint)
                            break;
                        SpBarArgument barArg = (SpBarArgument)line.Argument;
						Int64 var = eEvaluator.GetInteger(barArg.Terms[0]);
						Int64 max = eEvaluator.GetInteger(barArg.Terms[1]);
						Int64 length = eEvaluator.GetInteger(barArg.Terms[2]);
						console.Print(createBar(var, max, length));
					}
					if (line.FunctionCode == FunctionCode.BARL)
						console.NewLine();
					break;
				case FunctionCode.TIMES://小数計算。TIMES (変数) , (小数値)という形で使う。
					{
						SpTimesArgument timesArg = (SpTimesArgument)line.Argument;
						VariableToken var = timesArg.VariableDest;
						double d = eEvaluator.GetInteger(var) * timesArg.DoubleValue;
						FixedVariablePointer p = eEvaluator.GetFixedVariable(var);
						unchecked
						{
							vEvaluator.SetValue(p, (Int64)d);
						}
					}
					break;
				case FunctionCode.PRINT_ABL://能力。引数は登録番号
				case FunctionCode.PRINT_TALENT://素質
				case FunctionCode.PRINT_MARK://刻印
				case FunctionCode.PRINT_EXP://経験
					{
                        if (SkipPrint)
                            break;
                        ExpressionArgument intExpArg = (ExpressionArgument)line.Argument;
						Int64 target = eEvaluator.GetInteger(intExpArg.Term);
						console.Print(vEvaluator.GetCharacterDataString(target, line.FunctionCode));
						console.NewLine();
					}
					break;
				case FunctionCode.PRINT_PALAM://パラメータ
					{
                        if (SkipPrint)
                            break;
                        ExpressionArgument intExpArg = (ExpressionArgument)line.Argument;
						Int64 target = eEvaluator.GetInteger(intExpArg.Term);
						int count = 0;
						///100以降は否定の珠とかなので表示しない
						for (int i = 0; i < 100; i++)
						{
							string printStr = vEvaluator.GetCharacterParamString(target, i);
							if (printStr != null)
							{
								console.PrintC(printStr, true);
								count++;
								if ((Config.Instance.PrintCPerLine > 0) && (count % Config.Instance.PrintCPerLine == 0))
									console.PrintFlush(false);
							}
						}
						console.PrintFlush(false);
					}
					break;
				case FunctionCode.PRINT_ITEM://所持アイテム
                    if (SkipPrint)
                        break;
                    console.Print(vEvaluator.GetHavingItemsString());
					console.NewLine();
					break;
				case FunctionCode.PRINT_SHOPITEM://ショップで売っているアイテム
					{
                        if (SkipPrint)
                            break;
						int length = Math.Min(vEvaluator.ITEMSALES.Length, vEvaluator.ITEMNAME.Length);
						int count = 0;
						for (int i = 0; i < length; i++)
						{
							if (vEvaluator.ItemSales(i))
							{
								string printStr = vEvaluator.ITEMNAME[i];
								if (printStr == null)
									printStr = "";
								Int64 price = vEvaluator.ITEMPRICE[i];
								// 1.52a改変部分　（単位の差し替えおよび前置、後置に対応）
								if (Config.Instance.MoneyFirst)
									console.PrintC(string.Format("[{2}] {0}({3}{1})", printStr, price, i, Config.Instance.MoneyLabel), false);
								else
									console.PrintC(string.Format("[{2}] {0}({1}{3})", printStr, price, i, Config.Instance.MoneyLabel), false);
								count++;
								if ((Config.Instance.PrintCPerLine > 0) && (count % Config.Instance.PrintCPerLine == 0))
									console.PrintFlush(false);
							}
						}
						console.PrintFlush(false);
					}
					break;
				case FunctionCode.UPCHECK://パラメータの変動
                    vEvaluator.UpdateInUpcheck(console, SkipPrint);
					break;
                case FunctionCode.CUPCHECK://パラメータの変動(任意キャラ版)
                    {
                        ExpressionArgument intExpArg = (ExpressionArgument)line.Argument;
                        Int64 target = eEvaluator.GetInteger(intExpArg.Term);
                        vEvaluator.CUpdateInUpcheck(console, target, SkipPrint);
                    }
                    break;
                case FunctionCode.ADDCHARA://(キャラ番号)のキャラクタを追加
				case FunctionCode.ADDSPCHARA://(キャラ番号)のSPキャラクタを追加（フラグ0が非0であるcsvを探して追加）
				case FunctionCode.DELCHARA://(キャラ登録番号)のキャラクタを削除。
					{
						ExpressionArrayArgument intExpArg = (ExpressionArrayArgument)line.Argument;
						Int64 integer = -1;
						//for (int i = 0; i < intExpArg.TermList.Length; i++)
						foreach (IOperandTerm int64Term in intExpArg.TermList)
						{
							integer = eEvaluator.GetInteger(int64Term);
							if (line.FunctionCode == FunctionCode.ADDCHARA)
								vEvaluator.AddCharacter(integer, false);
							else if (line.FunctionCode == FunctionCode.ADDSPCHARA)
								vEvaluator.AddCharacter(integer, true);
							else
								vEvaluator.DelCharacter(integer);
						}
					}
					break;
				case FunctionCode.DELALLCHARA:
					{
						vEvaluator.DelAllCharacter();
						break;
					}
                case FunctionCode.PICKUPCHARA:
                    {
                        ExpressionArrayArgument intExpArg = (ExpressionArrayArgument)line.Argument;
                        Int64[] NoList = new Int64[intExpArg.TermList.Length];
                        int charaNum = (int)eEvaluator.VEvaluator.GetInteger(eEvaluator.GetFixedVariable(new VariableToken(VariableIdentifier.GetVariableId("CHARANUM"), new SingleTerm(0), new SingleTerm(0), new SingleTerm(0))));
                        for (int i = 0; i < intExpArg.TermList.Length; i++)
                        {
                            IOperandTerm term_i = intExpArg.TermList[i];
                            NoList[i] = eEvaluator.GetInteger(term_i);
                            if (!(term_i is VariableToken) || ((((VariableToken)term_i).Identifier.Code != VariableCode.MASTER) && (((VariableToken)term_i).Identifier.Code != VariableCode.ASSI) && (((VariableToken)term_i).Identifier.Code != VariableCode.TARGET)))
                            if (NoList[i] < 0 || NoList[i] >= charaNum)
                                throw new CodeEE("命令PICKUPCHARAの第" + (i + 1).ToString() + "引数にキャラリストの範囲外の値(" + NoList[i].ToString() + ")が与えられました");
                        }
                        vEvaluator.PickUpChara(NoList);
                    }
                    break;
                case FunctionCode.ADDDEFCHARA:
                    {
                        if (state.Scope != "SYSTEM_TITLE")
                            throw new CodeEE("@SYSTEM_TITLE以外でこの命令を使うことはできません");
                        vEvaluator.AddCharacterFromCsvNo(0);
                        if (Process.instance.gamebase.DefaultCharacter > 0)
                            vEvaluator.AddCharacterFromCsvNo(Process.instance.gamebase.DefaultCharacter);
                        break;
                    }
                case FunctionCode.ADDVOIDCHARA:
                    {
                        vEvaluator.AddPseudoCharacter();
                    }
                    break;
				case FunctionCode.PUTFORM://@SAVEINFO関数でのみ使用可能。PRINTFORMと同様の書式でセーブデータに概要をつける。
					{
						term = ((ExpressionArgument)line.Argument).Term;
						str = eEvaluator.GetString(term);
						if (vEvaluator.SAVEDATA_TEXT != null)
							vEvaluator.SAVEDATA_TEXT += str;
						else
							vEvaluator.SAVEDATA_TEXT = str;
						break;
					}
				case FunctionCode.QUIT://ゲームを終了
					console.Quit();
					break;


				case FunctionCode.STRLEN://RESULTに文字数（バイト数）を代入
				//case FunctionCode.STRLENS:
				case FunctionCode.STRLENFORM:
					term = ((ExpressionArgument)line.Argument).Term;
					str = eEvaluator.GetString(term);
					this.vEvaluator.RESULT = ShiftJisManager.GetStrlenShiftJis(str);
					break;

                case FunctionCode.STRLENU://RESULTに文字数（バイト数）を代入
                case FunctionCode.STRLENFORMU:
                    term = ((ExpressionArgument)line.Argument).Term;
                    str = eEvaluator.GetString(term);
                    this.vEvaluator.RESULT = str.Length;
                    break;

				case FunctionCode.SWAPCHARA:
					{
						SpSwapCharaArgument arg = (SpSwapCharaArgument)line.Argument;
						long x = eEvaluator.GetInteger(arg.X);
						long y = eEvaluator.GetInteger(arg.Y);
						vEvaluator.SwapChara(x, y);
					}
					break;
				case FunctionCode.SORTCHARA:
					{
						SpSortcharaArgument spSortArg = (SpSortcharaArgument)line.Argument;
						Int64 elem = 0;
						if (spSortArg.SortKey.Element2 != null)
							elem = eEvaluator.GetInteger(spSortArg.SortKey.Element2);
						vEvaluator.SortChara(spSortArg.SortKey.Identifier, elem, spSortArg.SortOrder, true);
					}
					break;
				case FunctionCode.VARSIZE:
					{
						SpVarsizeArgument versizeArg = (SpVarsizeArgument)line.Argument;
						VariableIdentifier varID = versizeArg.VariableID;
						vEvaluator.VarSize(varID);
					}
					break;
				case FunctionCode.SAVEDATA:
					{
						SpSaveDataArgument spSavedataArg = (SpSaveDataArgument)line.Argument;
						Int64 target = eEvaluator.GetInteger(spSavedataArg.Target);
						if (target < 0)
							throw new CodeEE("SAVEDATAの引数に負の値(" + target.ToString() + ")が指定されました");
						else if (target > int.MaxValue)
							throw new CodeEE("SAVEDATAの引数(" + target.ToString() + ")が大きすぎます");
						string savemes = eEvaluator.GetString(spSavedataArg.StrExpression);
                        if (savemes.Contains("\n"))
                            throw new CodeEE("SAVEDATAのセーブテキストに改行文字が与えられました（セーブデータが破損するため改行文字は使えません）");
						saveTo((int)target, savemes);
					}
					break;
				case FunctionCode.DELDATA:
					{
						ExpressionArgument intExpArg = (ExpressionArgument)line.Argument;
						Int64 target = eEvaluator.GetInteger(intExpArg.Term);
						if (target < 0)
							throw new CodeEE("DELDATAの引数に負の値(" + target.ToString() + ")が指定されました");
						else if (target > int.MaxValue)
							throw new CodeEE("DELDATAの引数(" + target.ToString() + ")が大きすぎます");
						delData((int)target);
					}
					break;
                //case FunctionCode.CHKDATA:
                //    {
                //        ExpressionArgument intExpArg = (ExpressionArgument)line.Argument;
                //        Int64 target = eEvaluator.GetInteger(intExpArg.Term);
                //        if (target < 0)
                //            throw new CodeEE("CHKDATAの引数に負の値(" + target.ToString() + ")が指定されました");
                //        else if (target > int.MaxValue)
                //            throw new CodeEE("CHKDATAの引数(" + target.ToString() + ")が大きすぎます");
                //        EraDataResult result = vEvaluator.checkData((int)target);
                //        vEvaluator.RESULT = (long)result.State;
                //        vEvaluator.RESULTS = result.DataMes;
                //    }
                //    break;
				case FunctionCode.SAVEGLOBAL:
					saveGlobal();
					break;
				case FunctionCode.LOADGLOBAL:
					if (loadGlobal())
						vEvaluator.RESULT = 1;
					else
						vEvaluator.RESULT = 0;
					break;
				case FunctionCode.POWER:
					{
						SpPowerArgument powerArg = (SpPowerArgument)line.Argument;
						FixedVariablePointer p = eEvaluator.GetFixedVariable(powerArg.VariableDest);
						double x = eEvaluator.GetInteger(powerArg.X);
						double y = eEvaluator.GetInteger(powerArg.Y);
						double pow = Math.Pow(x, y);
						if (double.IsNaN(pow))
							throw new CodeEE("累乗結果が非数値です");
						else if (double.IsInfinity(pow))
							throw new CodeEE("累乗結果が無限大です");
						else if ((pow >= Int64.MaxValue) || (pow <= Int64.MinValue))
							throw new CodeEE("累乗結果(" + pow.ToString() + ")が64ビット符号付き整数の範囲外です");
						vEvaluator.SetValue(p, (long)pow);
						break;
					}
				case FunctionCode.SWAP:
					{
						SpSwapVarArgument arg = (SpSwapVarArgument)line.Argument;
						FixedVariablePointer p1 = eEvaluator.GetFixedVariable(arg.var1);
						FixedVariablePointer p2 = eEvaluator.GetFixedVariable(arg.var2);
						if (arg.var1.GetOperandType() != arg.var2.GetOperandType())
							throw new CodeEE("入れ替える変数の型が異なります");
						if (arg.var1.GetOperandType() == typeof(Int64))
						{
                            Int64 temp = vEvaluator.GetInteger(p1);
                            vEvaluator.SetValue(p1, vEvaluator.GetInteger(p2));
							vEvaluator.SetValue(p2, temp);
						}
						else if (arg.var1.GetOperandType() == typeof(string))
						{
                            string temps = vEvaluator.GetString(p1);
                            vEvaluator.SetValue(p1, vEvaluator.GetString(p2));
							vEvaluator.SetValue(p2, temps);
						}
						else
						{
							throw new CodeEE("不明な変数型です");
						}
						break;
					}
				case FunctionCode.GETTIME:
					{
						long date = DateTime.Now.Year;
						date = date * 100 + DateTime.Now.Month;
						date = date * 100 + DateTime.Now.Day;
						date = date * 100 + DateTime.Now.Hour;
						date = date * 100 + DateTime.Now.Minute;
						date = date * 100 + DateTime.Now.Second;
						date = date * 1000 + DateTime.Now.Millisecond;
						vEvaluator.RESULT = date;//17桁。2京くらい。
                        vEvaluator.RESULTS = DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss");
					}
					break;
				case FunctionCode.SETCOLOR:
					{
						SpColorArgument colorArg = (SpColorArgument)line.Argument;
						Int64 colorR;
						Int64 colorG;
						Int64 colorB;
						if (colorArg.RGB != null)
						{
							Int64 colorRGB = eEvaluator.GetInteger(colorArg.RGB);
							colorR = (colorRGB & 0xFF0000) >> 16;
							colorG = (colorRGB & 0x00FF00) >> 8;
							colorB = (colorRGB & 0x0000FF);
						}
						else
						{
							colorR = eEvaluator.GetInteger(colorArg.R);
							colorG = eEvaluator.GetInteger(colorArg.G);
							colorB = eEvaluator.GetInteger(colorArg.B);
							if ((colorR < 0) || (colorG < 0) || (colorB < 0))
								throw new CodeEE("SETCOLORの引数に0未満の値が指定されました");
							if ((colorR > 255) || (colorG > 255) || (colorB > 255))
								throw new CodeEE("SETCOLORの引数に255を超える値が指定されました");
						}
						Color c = Color.FromArgb((Int32)colorR, (Int32)colorG, (Int32)colorB);
						console.SetStringStyle(c);
					}
					break;
				case FunctionCode.RESETCOLOR:
					console.SetStringStyle(Config.Instance.ForeColor);
					break;
				case FunctionCode.FONTBOLD:
					console.SetStringStyle(console.StringStyle.FontStyle | FontStyle.Bold);
					break;
				case FunctionCode.FONTITALIC:
					console.SetStringStyle(console.StringStyle.FontStyle | FontStyle.Italic);
					break;
				case FunctionCode.FONTREGULAR:
					console.SetStringStyle(FontStyle.Regular);
					break;
				case FunctionCode.FONTSTYLE:
					{
						FontStyle fs = FontStyle.Regular;
						Int64 i = eEvaluator.GetInteger(((ExpressionArgument)line.Argument).Term);
						if ((i & 1) != 0)
							fs |= FontStyle.Bold;
						if ((i & 2) != 0)
							fs |= FontStyle.Italic;
						if ((i & 4) != 0)
							fs |= FontStyle.Strikeout;
						if ((i & 8) != 0)
							fs |= FontStyle.Underline;
						console.SetStringStyle(fs);
					}
					break;
				case FunctionCode.SETFONT:
					term = ((ExpressionArgument)line.Argument).Term;
					console.SetFont(eEvaluator.GetString(term));
					break;
				case FunctionCode.ALIGNMENT:
					str = line.ArgumentStr;
					if (Config.Instance.IgnoreCase)
						str = str.ToUpper();
					if (!Enum.IsDefined(typeof(DisplayLineAlignment), str))
						throw new CodeEE("ALIGNMENTのキーワード\"" + str + "\"は未定義です");
					console.Alignment = (DisplayLineAlignment)Enum.Parse(typeof(DisplayLineAlignment), str);
					break;
				case FunctionCode.RESETDATA:
					vEvaluator.ResetData();
                    console.ResetStyle();
                    break;
				case FunctionCode.RESETGLOBAL:
					vEvaluator.ResetGlobalData();
					break;
				case FunctionCode.RANDOMIZE:
					vEvaluator.Randomize(eEvaluator.GetInteger(((ExpressionArgument)line.Argument).Term));
					break;
				case FunctionCode.INITRAND:
					vEvaluator.InitRanddata();
					break;
				case FunctionCode.DUMPRAND:
					vEvaluator.DumpRanddata();
					break;
				case FunctionCode.REDRAW:
					console.SetRedraw(eEvaluator.GetInteger(((ExpressionArgument)line.Argument).Term));
					break;
                case FunctionCode.SETBIT:
                case FunctionCode.CLEARBIT:
                case FunctionCode.INVERTBIT:
                    {
                        BitArgument spsetarg = (BitArgument)line.Argument;
                        FixedVariablePointer p = eEvaluator.GetFixedVariable(spsetarg.VariableDest);
                        Int64 x = eEvaluator.GetInteger(spsetarg.Term);
                        if ((x < 0) || (x > 63))
                            throw new CodeEE("第2引数がビットのレンジ(0から63)を超えています");
                        Int64 baseValue = vEvaluator.GetInteger(p);
                        Int64 shift = 1L << (int)x;
						if (line.FunctionCode == FunctionCode.SETBIT)
                            baseValue |= shift;
						else if (line.FunctionCode == FunctionCode.CLEARBIT)
                            baseValue &= ~shift;
                        else
                            baseValue ^= shift;
                        vEvaluator.SetValue(p, baseValue);
                    }
                    break;
				case FunctionCode.VARSET:
					{
						SpVarSetArgument spvarsetarg = (SpVarSetArgument)line.Argument;
						FixedVariablePointer p = eEvaluator.GetFixedVariable(spvarsetarg.VariableDest);
                        int start = 0;
                        if (spvarsetarg.Start != null)
                            start = (int)eEvaluator.GetInteger(spvarsetarg.Start);
                        int end = 0;
                        if (spvarsetarg.End != null)
                            end = (int)eEvaluator.GetInteger(spvarsetarg.End);
                        else if (spvarsetarg.VariableDest.Identifier.IsArray1D)
                            end = (int)eEvaluator.VEvaluator.GetVariableSize(p.ID);
                        if (start > end)
                        {
                            int temp = start;
                            start = end;
                            end = start;
                        }
                        if (p.ID.IsString)
						{
							string src = eEvaluator.GetString(spvarsetarg.Term);
							vEvaluator.SetValueAll(p, src, start, end);
						}
						else
						{
							long src = eEvaluator.GetInteger(spvarsetarg.Term);
                            vEvaluator.SetValueAll(p, src, start, end);
						}
					}
					break;
                case FunctionCode.CVARSET:
                    {
                        SpCVarSetArgument spvarsetarg = (SpCVarSetArgument)line.Argument;
                        FixedVariablePointer p = eEvaluator.GetFixedVariable(spvarsetarg.VariableDest);
                        SingleTerm index = eEvaluator.GetValue(spvarsetarg.Index);
                        int charaNum = (int)eEvaluator.VEvaluator.GetInteger(eEvaluator.GetFixedVariable(new VariableToken(VariableIdentifier.GetVariableId("CHARANUM"), new SingleTerm(0), new SingleTerm(0), new SingleTerm(0))));
                        int start = 0;
                        if (spvarsetarg.Start != null)
                        {
                            start = (int)eEvaluator.GetInteger(spvarsetarg.Start);
                            if (start < 0 || start >= charaNum)
                                throw new CodeEE("命令CVARSETの第４引数(" + start.ToString() + ")がキャラクタの範囲外です");
                        }
                        int end = 0;
                        if (spvarsetarg.End != null)
                        {
                            end = (int)eEvaluator.GetInteger(spvarsetarg.End);
                            if (end < 0 || end > charaNum)
                                throw new CodeEE("命令CVARSETの第５引数(" + end.ToString() + ")がキャラクタの範囲外です");
                        }
                        else
                            end = charaNum;
                        if (start > end)
                        {
                            int temp = start;
                            start = end;
                            end = start;
                        }
                        if (!p.ID.IsCharacterData)
                            throw new CodeEE("命令CVARSETにキャラクタ変数でない変数" + p.ID.ToString() + "が渡されました");
                        if (index.GetOperandType() == typeof(string) && p.ID.IsArray1D)
                        {
                            if (!ConstantData.isDefined(p.ID.Code, index.Str))
                                throw new CodeEE("文字列" + index.Str + "は配列変数" + p.ID.Code.ToString() + "の要素ではありません");
                        }
                        if (p.ID.IsString)
                        {
                            string src = eEvaluator.GetString(spvarsetarg.Term);
                            vEvaluator.SetValueAllEachChara(p, index, src, start, end);
                        }
                        else
                        {
                            long src = eEvaluator.GetInteger(spvarsetarg.Term);
                            vEvaluator.SetValueAllEachChara(p, index, src, start, end);
                        }
                    }
                    break;
                case FunctionCode.RESET_STAIN:
                    {
                        ExpressionArgument expArg = (ExpressionArgument)line.Argument;
                        Int64 no = eEvaluator.GetInteger(expArg.Term);
                        vEvaluator.setDefaultStain(no);
                    }
                    break;
                case FunctionCode.SPLIT:
                    {
                        SpSplitArgument spSplitArg = (SpSplitArgument)line.Argument;
                        FixedVariablePointer p;
                        string target = eEvaluator.GetString(spSplitArg.TargetStr);
                        string[] split = new string[] { eEvaluator.GetString(spSplitArg.Split) };
                        string[] retStr = target.Split(split, StringSplitOptions.None);
                        for (int i = 0; i < retStr.Length; i++)
                        {
                            p = eEvaluator.GetFixedVariable(VariableParser.ReduceVariable(spSplitArg.Var, new SingleTerm(i), null, null));
                            vEvaluator.SetValue(p, retStr[i]);
                        }
                        vEvaluator.RESULT = retStr.Length;
                    }
                    break;
                case FunctionCode.PRINTCPERLINE:
                    {
                        SpGetIntArgument spGetintArg = (SpGetIntArgument)line.Argument;
                        FixedVariablePointer p = eEvaluator.GetFixedVariable(spGetintArg.VarToken);
                        vEvaluator.SetValue(p, (Int64)Config.Instance.PrintCPerLine);
                    }
                    break;
                case FunctionCode.SAVENOS:
                    {
                        SpGetIntArgument spGetintArg = (SpGetIntArgument)line.Argument;
                        FixedVariablePointer p = eEvaluator.GetFixedVariable(spGetintArg.VarToken);
                        vEvaluator.SetValue(p, (Int64)Config.Instance.SaveDataNos);
                    }
                    break;
                case FunctionCode.FORCEKANA:
                    {
                        ExpressionArgument arg = (ExpressionArgument)line.Argument;
                        Int64 flag = eEvaluator.GetInteger(arg.Term);
                        if (flag < 0 || flag > 3)
                            throw new CodeEE("命令FORCEKANAの引数が指定可能な範囲(0～3)を超えています");
                        forceKatakana = (flag == 1) ? true : false;
                        forceHiragana = (flag > 1) ? true : false;
                        halftoFull = (flag == 3) ? true : false;
                    }
                    break;
                case FunctionCode.CALLF:
                    {
                        string label = ((SpCallArgment)line.Argument).Str;
                        if (Config.Instance.IgnoreCase)
                            label = label.ToUpper();
                        IOperandTerm mToken = FunctionMethodCreator.GetFunctionMethod(labelDic, label, ((SpCallArgment)line.Argument).Args);
                        if (mToken == null)
                            throw new CodeEE("式中関数\"@" + label + "\"が見つかりません");
                        eEvaluator.GetValue(mToken);
                        break;
                    }
                case FunctionCode.CALLFORMF:
                    {
                        SpCallformArgment spCallformArg = (SpCallformArgment)line.Argument;
                        StringForm strForm = spCallformArg.StrForm;
                        string label = strForm.GetString(eEvaluator).Trim();
                        if (Config.Instance.IgnoreCase)
                            label = label.ToUpper();
                        IOperandTerm mToken = FunctionMethodCreator.GetFunctionMethod(labelDic, label, spCallformArg.Args);
                        if (mToken == null)
                            throw new CodeEE("式中関数\"@" + label + "\"が見つかりません");
                        eEvaluator.GetValue(mToken);
                        break;
                    }
                case FunctionCode.SKIPDISP:
                    {
                        ExpressionArgument arg = (ExpressionArgument)line.Argument;
                        Int64 skip = eEvaluator.GetInteger(arg.Term);
                        if (skip != 0)
                        {
                            SkipPrint = true;
                            userDefinedSkip = true;
                        }
                        else
                        {
                            SkipPrint = false;
                            userDefinedSkip = false;
                        }
                        vEvaluator.RESULT = eEvaluator.GetInteger(new SingleTerm(SkipPrint));
                    }
                    break;
                case FunctionCode.NOSKIP:
                    {
                        if (line.JumpTo == null)
                            throw new CodeEE("対応するENDNOSKIPのないNOSKIPです");
                        if (skipPrint)
                        {
                            saveSkip = true;
                            skipPrint = false;
                        }
                        else
                            saveSkip = false;
                    }
                    break;
                case FunctionCode.ENDNOSKIP:
                    {
                        if (line.JumpTo == null)
                            throw new CodeEE("対応するNOSKIPのないENDNOSKIPです");
                        if (saveSkip)
                            skipPrint = true;
                    }
                    break;
                case FunctionCode.OUTPUTLOG:
                    console.outputLog(null);
                    break;
                case FunctionCode.ARRAYSHIFT: //配列要素をずらす
                    {
                        SpArrayShiftArgument arrayArg = (SpArrayShiftArgument)line.Argument;
                        if (!arrayArg.VarToken.Identifier.IsArray1D)
                            throw new CodeEE("ARRAYSHIFTは1次元配列および配列型キャラクタ変数のみに対応しています");
                        FixedVariablePointer dest = eEvaluator.GetFixedVariable(arrayArg.VarToken);
                        int shift = (int)eEvaluator.GetInteger(arrayArg.Num1);
                        if (shift == 0)
                            break;
                        int start = (int)eEvaluator.GetInteger(arrayArg.Num3);
                        if (start <  0)
                            throw new CodeEE("ARRAYSHIFTの第３引数が負の値(" + start.ToString() + ")です");
                        int num;
                        if (arrayArg.Num4 != null)
                        {
                            num = (int)eEvaluator.GetInteger(arrayArg.Num4);
                            if (num < 0)
                                throw new CodeEE("ARRAYSHIFTの第４引数が負の値(" + start.ToString() + ")です");
                            if (num == 0)
                                break;
                        }
                        else
                            num = -1;
                        if (dest.ID.IsInteger)
                        {
                            Int64 def = eEvaluator.GetInteger(arrayArg.Num2);
                            vEvaluator.ShiftArray(dest, shift, def, start, num);
                        }
                        else
                        {
                            string defs = eEvaluator.GetString(arrayArg.Num2);
                            vEvaluator.ShiftArray(dest, shift, defs, start, num);
                        }
                        break;
                    }
                case FunctionCode.ARRAYREMOVE:
                    {
                        SpArrayControlArgument arrayArg = (SpArrayControlArgument)line.Argument;
                        if (!arrayArg.VarToken.Identifier.IsArray1D)
                            throw new CodeEE("ARRAYREMOVEは1次元配列および配列型キャラクタ変数のみに対応しています");
                        FixedVariablePointer p = eEvaluator.GetFixedVariable(arrayArg.VarToken);
                        int start = (int)eEvaluator.GetInteger(arrayArg.Num1);
                        int num = (int)eEvaluator.GetInteger(arrayArg.Num2);
                        if (start < 0)
                            throw new CodeEE("ARRAYREMOVEの第１引数が負の値(" + start.ToString() + ")です");
                        vEvaluator.RemoveArray(p, start, num);
                        break;
                    }
                case FunctionCode.ENCODETOUNI:
                    {
                        term = ((ExpressionArgument)line.Argument).Term;
                        string target = eEvaluator.GetString(term);
                        int length = Encoding.UTF32.GetEncoder().GetByteCount(target.ToCharArray(), 0, target.Length, false);
                        byte[] bytes = new byte[length];
                        Encoding.UTF32.GetEncoder().GetBytes(target.ToCharArray(), 0, target.Length, bytes, 0, false);
                        vEvaluator.setEncodingResult(bytes);
                    }
					break;
				case FunctionCode.DEBUGPRINT:
				case FunctionCode.DEBUGPRINTL:
				case FunctionCode.DEBUGPRINTFORM:
				case FunctionCode.DEBUGPRINTFORML:
                    {
						term = ((ExpressionArgument)line.Argument).Term;
						str = eEvaluator.GetString(term);
						console.DebugPrint(str);
						if (line.Function.IsNewLine())
							console.DebugNewLine();
					}
					break;
				case FunctionCode.ASSERT:
					term = ((ExpressionArgument)line.Argument).Term;
					if (eEvaluator.GetInteger(term) == 0)
						throw new CodeEE("ASSERT文の引数が0です");
					break;
                default:
					throw new ExeEE("未定義の関数");
			}
			return;
		}

        bool saveSkip = false;
        bool userDefinedSkip = false;

        private void outputToConsole(string str, FunctionIdentifier func)
        {
			if (func.IsPrintSingle())
                console.PrintLine(str);
            else
            {
                console.Print(str);
				if (func.IsNewLine() || func.IsWaitInput())
                    console.NewLine();
				if (func.IsWaitInput())
                    console.ReadAnyKey();
            }
            console.UseUserStyle = true;
        }

        private string convertStringType(string str)
        {
            if (forceKatakana)
                return Strings.StrConv(str, VbStrConv.Katakana, 0);
            else if (forceHiragana)
            {
                if (halftoFull)
                    return Strings.StrConv(str, VbStrConv.Hiragana | VbStrConv.Wide, 0);
                else
                    return Strings.StrConv(str, VbStrConv.Hiragana, 0);
            }
            return str;
        }

		private string createBar(Int64 var, Int64 max, Int64 length)
		{
			if (max <= 0)
				throw new CodeEE("BARの最大値が正の値ではありません");
			if (length <= 0)
				throw new CodeEE("BARの長さが正の値ではありません");
			if (length >= 100)//暴走を防ぐため。
				throw new CodeEE("BARが長すぎます");
			StringBuilder builder = new StringBuilder();
			builder.Append('[');
			int count;
			unchecked
			{
				count = (int)(var * length / max);
			}
			if (count < 0)
				count = 0;
			if (count > length)
				count = (int)length;
			builder.Append(Config.Instance.BarChar1, count);
			builder.Append(Config.Instance.BarChar2, (int)length - count);
			builder.Append(']');
			return builder.ToString();
		}

        bool forceKatakana = false;
        bool forceHiragana = false;
        bool halftoFull = false;
		#endregion

		#region flow control

		bool doFlowControlFunction(InstructionLine func, bool sequential)
		{
            StringStream st = new StringStream(func.ArgumentStr);
            ExpressionArgument expArg = null;
            switch (func.FunctionCode)
            {
                case FunctionCode.BEGIN://システム関数の実行。
                    string keyword = func.ArgumentStr;
                    if (Config.Instance.IgnoreCase)
                        keyword = keyword.ToUpper();
                    if (!Enum.IsDefined(typeof(BeginType), keyword))
                        throw new CodeEE("BEGINのキーワード\"" + keyword + "\"は未定義です");
                    state.SetBegin((BeginType)Enum.Parse(typeof(BeginType), keyword));
                    state.Return(0);
                    console.ResetStyle();
                    return false;
                case FunctionCode.SAVEGAME://セーブ画面を呼ぶ。ショップのみ。
                    saveCurrentState(true);
                    state.SaveLoadData(true);
                    return false;
                case FunctionCode.LOADGAME://
                    saveCurrentState(true);
                    state.SaveLoadData(false);
                    return false;

                //case FunctionCode.SIF://一行のみIF
                //    expArg = (ExpressionArgument)func.Argument;
                //    if (eEvaluator.GetInteger(expArg.Term) != 0)//評価式が真ならそのまま流れ落ちる
                //        break;
                //    state.ShfitNextLine();//偽なら一行とばす。順に来たときと同じ扱いにする
                //    break;
                //case FunctionCode.IF:
                //    {//1.723 仕様変更
                //        LogicalLine ifJumpto = null;
                //        if (func.IfCaseList == null)
                //            throw new ExeEE("IFのIF-ELSEIFリストが適正に作成されていない");
                //        if (func.JumpTo == null)
                //            throw new ExeEE("IFに対応するENDIFが設定されていない");

                //        for (int i = 0; i < func.IfCaseList.Count; i++)
                //        {
                //            if (func.IfCaseList[i].IsError)
                //                continue;
                //            if (func.IfCaseList[i].Function == FunctionCode.ELSE)
                //            {
                //                ifJumpto = func.IfCaseList[i];
                //                break;
                //            }

                //            expArg = (ExpressionArgument)(func.IfCaseList[i].Argument);
                //            if (expArg == null)
                //                throw new ExeEE("IFチェック中。引数が解析されていない。", func.IfCaseList[i].Position);

                //            //1730 ELSEIFが出したエラーがIFのエラーとして検出されていた
                //            state.RunningLine = func.IfCaseList[i];
                //            if (eEvaluator.GetInteger(expArg.Term) != 0)//式が真
                //            {
                //                ifJumpto = func.IfCaseList[i];
                //                break;
                //            }
                //        }
                //        if (ifJumpto == null)
                //            ifJumpto = func.JumpTo;//ENDIF
                //        if (ifJumpto != func)//自分自身がジャンプ先ならそのまま
                //            state.JumpTo(ifJumpto);
                //        state.RunningLine = null;
                //    }
                //    break;
                //case FunctionCode.SELECTCASE:
                //    {//1.724 追加
                //        LogicalLine caseJumpto = null;
                //        SingleTerm term = eEvaluator.GetValue(((ExpressionArgument)func.Argument).Term);
                //        if (func.IfCaseList == null)
                //            throw new ExeEE("SELECTCASEのCASEリストが適正に作成されていない");
                //        if (func.JumpTo == null)
                //            throw new ExeEE("SELECTCASEに対応するENDSELECTが設定されていない");
                //        for (int i = 0; i < func.IfCaseList.Count; i++)
                //        {
                //            if (func.IfCaseList[i].IsError)
                //                continue;
                //            if (func.IfCaseList[i].Function == FunctionCode.CASEELSE)
                //            {
                //                caseJumpto = func.IfCaseList[i];
                //                break;
                //            }
                //            CaseArgument caseArg = (CaseArgument)(func.IfCaseList[i].Argument);
                //            if (caseArg == null)
                //                throw new ExeEE("CASEチェック中。引数が解析されていない。", func.IfCaseList[i].Position);

                //            //1730 CASEが出したエラーがSELECTCASEのエラーとして検出されていた
                //            state.RunningLine = func.IfCaseList[i];
                //            foreach (CaseExpression caseExp in caseArg.CaseExps)
                //            {
                //                if (eEvaluator.GetCase(caseExp, term))
                //                {
                //                    caseJumpto = func.IfCaseList[i];
                //                    goto casefound;
                //                }
                //            }
                //        }
                //    casefound:
                //        if (caseJumpto == null)
                //            caseJumpto = func.JumpTo;//ENDSELECT
                //        state.JumpTo(caseJumpto);
                //        state.RunningLine = null;
                //    }
                //    break;
                //case FunctionCode.ELSE:
                //case FunctionCode.ELSEIF:
                //case FunctionCode.CASE:
                //case FunctionCode.CASEELSE:
                //    if (sequential)//直接来たならENDIFへ
                //    {
                //        if (func.JumpTo == null)
                //            throw new ExeEE(func.Function.ToString() + "のジャンプ先が設定されていない");
                //        state.JumpTo(func.JumpTo);
                //        break;
                //    }
                    //else//IFから飛ばされたならそのまま次へ
                    //{//1.723 ELSEIFの条件判定はIFで行う。
                    //    break;
                    //}
                //case FunctionCode.ENDIF:
                //case FunctionCode.ENDSELECT:
                //case FunctionCode.ENDCATCH:
                //case FunctionCode.ENDFUNC:
                //case FunctionCode.DO:
                //    break;//ENDIF, ENDSELECTでするべきことは特に無い

                case FunctionCode.REPEAT://RENDまで繰り返し。繰り返した回数がCOUNTへ。ネスト不可。
                case FunctionCode.FOR://NEXTまで繰り返し。カウンター変数を設定可能。ネスト可。
                    //if (sequential)//上から来たなら回数のセット
                    //{
                    if (func.FunctionCode == FunctionCode.FOR)
                    {
                        SpForNextArgment forArg = (SpForNextArgment)func.Argument;
                        func.LoopCounter = eEvaluator.GetFixedVariable(forArg.Cnt);
                        //1.725 順序変更。REPEATにならう。
                        vEvaluator.SetValue(func.LoopCounter, eEvaluator.GetInteger(forArg.Start));
                        func.LoopEnd = eEvaluator.GetInteger(forArg.End);
                        func.LoopStep = eEvaluator.GetInteger(forArg.Step);
                    }
                    else
                    {
                        expArg = (ExpressionArgument)func.Argument;
                        FixedVariablePointer count = new FixedVariablePointer();
                        count.ID = VariableIdentifier.GetVariableId(VariableCode.COUNT);
                        count.Index1 = 0;
                        //1.725 順序変更。eramakerでは引数評価の前にCOUNTに0を代入していた。
                        vEvaluator.SetValue(count, 0);
                        func.LoopEnd = eEvaluator.GetInteger(expArg.Term);
                        func.LoopCounter = count;
                        func.LoopStep = 1;
                    }
                    //}
                    //else//（REND、CONTINUEから）飛んできたならカウンターの増加
                    //{
                    //    unchecked
                    //    {
                    //        vEvaluator.SetValue(func.LoopCounter, vEvaluator.GetInteger(func.LoopCounter) + func.LoopStep);
                    //    }
                    //}
                    if ((func.LoopStep > 0) && (func.LoopEnd > vEvaluator.GetInteger(func.LoopCounter)))//まだ回数が残っているなら、
                        break;//そのまま次の行へ
                    else if ((func.LoopStep < 0) && (func.LoopEnd < vEvaluator.GetInteger(func.LoopCounter)))//まだ回数が残っているなら、
                        break;//そのまま次の行へ
                    state.JumpTo(func.JumpTo);
                    break;
                case FunctionCode.WHILE://WENDまで繰り返し、与えられた条件を満たす限り無限に繰り返す。多分ネスト可
                    {
                        expArg = (ExpressionArgument)func.Argument;
                        if (eEvaluator.GetInteger(expArg.Term) != 0)//式が真
                            break;//そのまま中の処理へ
                        state.JumpTo(func.JumpTo);
                        break;
                    }
                //case FunctionCode.NEXT://NEXTに戻る
                //case FunctionCode.REND://REPEATに戻る
                //    {
                //        InstructionLine jumpLine = (InstructionLine)func.JumpTo;
                //        unchecked
                //        {
                //            vEvaluator.SetValue(jumpLine.LoopCounter, vEvaluator.GetInteger(jumpLine.LoopCounter) + jumpLine.LoopStep);
                //        }
                //        if (((jumpLine.LoopStep > 0) && (jumpLine.LoopEnd > vEvaluator.GetInteger(jumpLine.LoopCounter)))
                //            || ((jumpLine.LoopStep < 0) && (jumpLine.LoopEnd < vEvaluator.GetInteger(jumpLine.LoopCounter))))//まだ回数が残っているなら、
                //            state.JumpTo(func.JumpTo);
                //    }
                //    break;
                //case FunctionCode.WEND://WHILEに戻る
                //    if (!sequential)//REPEAT、FOR、WHILE、BREAKから飛んできたなら
                //        break;
                //    //順列で来たならREPEAT、FOR、WHILEに帰る。繰り返し回数の記憶とチェックはREPEAT、FORが行う。
                //    state.JumpTo(func.JumpTo);
                //    break;
                //case FunctionCode.BREAK://RENDまでジャンプ
                //    {
                //        ////BREAKのJUMP先はRENDまたはNEXT。そのジャンプ先であるREPEATかFORをiLineに代入。
                //        //1.723 仕様変更。BREAKのJUMP先にはREPEAT、FOR、WHILEを記憶する。そのJUMP先が本当のJUMP先。
                //        InstructionLine iJumpTo = (InstructionLine)func.JumpTo;
                //        InstructionLine iLine = (InstructionLine)iJumpTo.JumpTo;
                //        //WHILEとDOはカウンタがないので、即ジャンプ
                //        if (iJumpTo.Function == FunctionCode.WHILE || iJumpTo.Function == FunctionCode.DO)
                //        {
                //            state.JumpTo(iLine);
                //            break;
                //        }
                //        unchecked
                //        {//本家ではBREAK時にCOUNTが回る
                //            vEvaluator.SetValue(iJumpTo.LoopCounter, vEvaluator.GetInteger(iJumpTo.LoopCounter) + iJumpTo.LoopStep);
                //        }
                //        state.JumpTo(iLine);
                //        break;
                //    }
                //case FunctionCode.CONTINUE://REPEATに戻る
                //    {
                //        InstructionLine jumpTo = (InstructionLine)func.JumpTo;
                //        if (jumpTo.Function == FunctionCode.WEND)
                //        {
                //            expArg = (ExpressionArgument)jumpTo.Argument;
                //            if (eEvaluator.GetInteger(expArg.Term) != 0)//式が真
                //                state.JumpTo(func.JumpTo);
                //            else
                //                state.JumpTo(jumpTo.JumpTo);
                //        }
                //        else
                //        {
                //            unchecked
                //            {
                //                vEvaluator.SetValue(jumpTo.LoopCounter, vEvaluator.GetInteger(jumpTo.LoopCounter) + jumpTo.LoopStep);
                //            }
                //            if (((jumpTo.LoopStep > 0) && (jumpTo.LoopEnd > vEvaluator.GetInteger(jumpTo.LoopCounter)))
                //                || ((jumpTo.LoopStep < 0) && (jumpTo.LoopEnd < vEvaluator.GetInteger(jumpTo.LoopCounter))))//まだ回数が残っているなら、
                //                state.JumpTo(func.JumpTo);
                //            else
                //                state.JumpTo(jumpTo.JumpTo);
                //        }
                //    }
                //    break;
                case FunctionCode.GOTO://$ラベルへジャンプ
                case FunctionCode.TRYGOTO:
                case FunctionCode.TRYCGOTO:
                //case FunctionCode.RESTART://関数の再開。関数の最初に戻る。
                    if (func.JumpTo == null)
                    {
						if (func.FunctionCode == FunctionCode.TRYGOTO)
                            break;
						else if (func.FunctionCode == FunctionCode.TRYCGOTO)
                        {
                            if (func.JumpToEndCatch != null)
                                state.JumpTo(func.JumpToEndCatch);
                            break;
                        }
						if (func.FunctionCode == FunctionCode.GOTO)
                            throw new CodeEE("ラベル\"$" + func.ArgumentStr + "\"が見つかりません");
                        throw new ExeEE("ジャンプ先が適切に設定されていない");
                    }
                    state.JumpTo(func.JumpTo);
                    break;
                case FunctionCode.JUMP://関数に移動
                case FunctionCode.TRYJUMP:
                case FunctionCode.TRYCJUMP:
                    {
                        if (func.JumpTo == null)
                        {
                            if (func.FunctionCode == FunctionCode.TRYJUMP)
                                break;
                            else if (func.FunctionCode == FunctionCode.TRYCJUMP)
                            {
                                if (func.JumpToEndCatch != null)
                                    state.JumpTo(func.JumpToEndCatch);
                                break;
                            }
                            throw new CodeEE("関数\"@" + ((SpCallArgment)func.Argument).Str + "\"が見つかりません");
                        }
                        assignArgs((FunctionLabelLine)func.JumpTo, ((SpCallArgment)func.Argument).Args);
                        string label = ((SpCallArgment)func.Argument).Str;
                        CalledFunction call = CalledFunction.CallFunction(this, label, state.GetCurrentReturnAddress, false);
                        state.ChangeCurrentFunction(call);
                        state.JumpTo(func.JumpTo);
                    }
                    break;
                case FunctionCode.CALL://関数に移動。移動元を記憶し、RETURNで帰る。
                case FunctionCode.TRYCALL:
                case FunctionCode.TRYCCALL:
                    {
                        if ((func.JumpTo == null) && (func.FunctionCode == FunctionCode.TRYCALL))
                            break;
                        else if ((func.JumpTo == null) && (func.FunctionCode == FunctionCode.TRYCCALL))
                        {
                            if (func.JumpToEndCatch != null)
                                state.JumpTo(func.JumpToEndCatch);
                            break;
                        }
                        if (!sequential)//RETURNで帰ってきた
                            break;//何もしない
                        string label = ((SpCallArgment)func.Argument).Str;
                        if (Config.Instance.IgnoreCase)
                            label = label.ToUpper();
                        CalledFunction call = CalledFunction.CallFunction(this, label, func, false);
                        if ((call == null) || (call.Count == 0) || (call.NextLine == null))
                            throw new CodeEE("関数\"@" + label + "\"が見つかりません");
                        if (call.Count > 1)
                            throw new ExeEE("関数\"@" + label + "\"の候補が複数返された");
                        //if (call.LabelList.Count > 1 && strArray.Count > 0)
                        //    throw new ExeEE("EVENT関数\"@" + label + "\"に引数を指定することはできません");
                        assignArgs(call.LabelList[0], ((SpCallArgment)func.Argument).Args);
                        state.AddFunction(call);
                        break;
                    }
                //case FunctionCode.RETURN://関数の終了。RESULTに整数を格納可能。
                //    {
                //        expArg = (ExpressionArgument)func.Argument;
                //        Int64 ret = eEvaluator.GetInteger(expArg.Term);
                //        vEvaluator.RESULT = ret;
                //        state.Return(ret);
                //        if (state.GetCurrentReturnAddress is NullLine)
                //            return false;
                //        break;
                //    }
                //case FunctionCode.RETURNFORM://関数の終了。RESULTに整数を格納可能。
                //    {
                //        string arg = eEvaluator.GetString(((ExpressionArgument)func.Argument).Term);
                //        StringStream aSt = new StringStream(arg);
                //        Int64 ret = eEvaluator.GetInteger(ExpressionParser.ReduceIntegerTerm(aSt, null));
                //        vEvaluator.RESULT = ret;
                //        state.Return(ret);
                //        if (state.GetCurrentReturnAddress is NullLine)
                //            return false;
                //        break;
                //    }
                case FunctionCode.RETURNF:
                    {
                        expArg = (ExpressionArgument)func.Argument;
                        SingleTerm ret = null;
                        if (expArg.Term != null)
                        {
                            ret = eEvaluator.GetValue(expArg.Term);
                        }
                        state.ReturnF(ret);
                        return false;
                    }
                case FunctionCode.LOADDATA:
                    {
                        ExpressionArgument intExpArg = (ExpressionArgument)func.Argument;
                        Int64 target = eEvaluator.GetInteger(intExpArg.Term);
                        if (target < 0)
                            throw new CodeEE("LOADDATAの引数に負の値(" + target.ToString() + ")が指定されました");
                        else if (target > int.MaxValue)
                            throw new CodeEE("LOADDATAの引数(" + target.ToString() + ")が大きすぎます");
                        EraDataResult result = vEvaluator.checkData((int)target);
                        if (result.State != EraDataState.OK)
                            throw new CodeEE("不正なデータをロードしようとしました");

                        loadFrom((int)target);
                        state.ClearFunctionList();
                        state.SystemState = SystemStateCode.LoadData_DataLoaded;
                        return false;
                    }
                case FunctionCode.JUMPFORM:
                case FunctionCode.TRYJUMPFORM:
                case FunctionCode.TRYCJUMPFORM:
                    {
                        SpCallformArgment spCallformArg = (SpCallformArgment)func.Argument;
                        StringForm strForm = spCallformArg.StrForm;
                        string label = strForm.GetString(eEvaluator).Trim();
                        if (Config.Instance.IgnoreCase)
                            label = label.ToUpper(); ;
                        CalledFunction call = CalledFunction.CallFunction(this, label, state.GetCurrentReturnAddress, false);
                        if ((call == null) || (call.Count == 0) || (call.NextLine == null))
                        {
                            if (func.FunctionCode == FunctionCode.TRYJUMPFORM)
                                break;
                            else if (func.FunctionCode == FunctionCode.TRYCJUMPFORM)
                            {
                                if (func.JumpToEndCatch != null)
                                    state.JumpTo(func.JumpToEndCatch);
                                break;
                            }
                            throw new CodeEE("関数\"@" + label + "\"が見つかりません");
                        }
                        assignArgs(call.LabelList[0], spCallformArg.Args);
                        state.ChangeCurrentFunction(call);
                        state.JumpTo(call.NextLine);
                        break;
                    }

                case FunctionCode.CALLFORM:
                case FunctionCode.TRYCALLFORM:
                case FunctionCode.TRYCCALLFORM:
                    {
                        if (!sequential)//RETURNで帰ってきた
                            break;//何もしない
                        SpCallformArgment spCallformArg = (SpCallformArgment)func.Argument;
                        StringForm strForm = spCallformArg.StrForm;
                        string label = strForm.GetString(eEvaluator).Trim();
                        if (Config.Instance.IgnoreCase)
                            label = label.ToUpper(); ;
                        CalledFunction call = CalledFunction.CallFunction(this, label, func, false);
                        if ((call == null) || (call.Count == 0) || (call.NextLine == null))
                        {
                            if (func.FunctionCode == FunctionCode.TRYCALLFORM)
                                break;
                            else if (func.FunctionCode == FunctionCode.TRYCCALLFORM)
                            {
                                if (func.JumpToEndCatch != null)
                                    state.JumpTo(func.JumpToEndCatch);
                                break;
                            }
                            throw new CodeEE("関数\"@" + label + "\"が見つかりません");

                        }
                        assignArgs(call.LabelList[0], spCallformArg.Args);
                        state.AddFunction(call);
                        break;
                    }
                case FunctionCode.GOTOFORM:
                case FunctionCode.TRYGOTOFORM:
                case FunctionCode.TRYCGOTOFORM:
                    {
                        string label = eEvaluator.GetString(((ExpressionArgument)func.Argument).Term);
                        if (Config.Instance.IgnoreCase)
                            label = label.ToUpper();
                        LogicalLine jumpto = state.CurrentCalled.CallLabel(this, label);
                        if (jumpto == null)
                        {
                            if (func.FunctionCode == FunctionCode.TRYGOTOFORM)
                                break;
                            else if (func.FunctionCode == FunctionCode.TRYCGOTOFORM)
                            {
                                if (func.JumpToEndCatch != null)
                                    state.JumpTo(func.JumpToEndCatch);
                                break;
                            }
                            throw new CodeEE("ラベル\"$" + label + "\"が見つかりません");
                        }
                        state.JumpTo(jumpto);
                        break;
                    }
                case FunctionCode.TRYCALLLIST:
                    {
                        if (!sequential)//RETURNで帰ってきた
                        {
                            state.JumpTo(func.JumpTo);
                            break;
                        }
                        string funcName = "";
                        CalledFunction callto = null;
                        SpCallformArgment cfa = null;
                        foreach (InstructionLine iLine in func.callList)
                        {
                            if (iLine.Argument == null)
                                LogicalLineParser.SetArgumentTo(iLine, null);
                            cfa = (SpCallformArgment)iLine.Argument;
                            funcName = cfa.StrForm.GetString(eEvaluator).Trim();
                            if (Config.Instance.IgnoreCase)
                                funcName = funcName.ToUpper();
                            callto = CalledFunction.CallFunction(this, funcName, func.JumpTo, false);
                            if ((callto != null) && (callto.Count != 0) && (callto.NextLine != null))
                                break;
                        }
                        if ((callto == null) || (callto.Count == 0) || (callto.NextLine == null))
                            state.JumpTo(func.JumpTo);
                        else
                        {
                            assignArgs(callto.LabelList[0], cfa.Args);
                            state.AddFunction(callto);
                        }
                    }
                    break;
                case FunctionCode.TRYJUMPLIST:
                    {
                        string funcName = "";
                        CalledFunction callto = null;
                        SpCallformArgment cfa = null;
                        foreach (InstructionLine iLine in func.callList)
                        {
                            if (iLine.Argument == null)
                                LogicalLineParser.SetArgumentTo(iLine, null);
                            cfa = (SpCallformArgment)iLine.Argument;
                            funcName = cfa.StrForm.GetString(eEvaluator).Trim();
                            if (Config.Instance.IgnoreCase)
                                funcName = funcName.ToUpper();
                            callto = CalledFunction.CallFunction(this, funcName, func, false);
                            if ((callto != null) && (callto.Count != 0) && (callto.NextLine != null))
                                break;
                        }
                        if ((callto == null) || (callto.Count == 0) || (callto.NextLine == null))
                            state.JumpTo(func.JumpTo);
                        else
                        {
                            assignArgs(callto.LabelList[0], cfa.Args);
                            state.ChangeCurrentFunction(callto);
                            state.JumpTo(callto.NextLine);
                        }
                    }
                    break;
                case FunctionCode.TRYGOTOLIST:
                    {
                        string funcName = "";
                        LogicalLine jumpto = null;
                        foreach (InstructionLine iLine in func.callList)
                        {
                            if (iLine.Argument == null)
                                LogicalLineParser.SetArgumentTo(iLine, null);
                            funcName = ((SpCallformArgment)iLine.Argument).StrForm.GetString(eEvaluator).Trim();
                            if (Config.Instance.IgnoreCase)
                                funcName = funcName.ToUpper();
                            jumpto = state.CurrentCalled.CallLabel(this, funcName);
                            if (jumpto != null)
                                break;
                        }
                        if (jumpto == null)
                            state.JumpTo(func.JumpTo);
                        else
                            state.JumpTo(jumpto);
                    }
                    break;
                case FunctionCode.CALLTRAIN:
                    {
                        ExpressionArgument intExpArg = (ExpressionArgument)func.Argument;
                        Int64 count = eEvaluator.GetInteger(intExpArg.Term);
                        SetCommnds(count);
                        return false;
                    }
                case FunctionCode.CATCH:
                    {
                        if (sequential)//上から流れてきたなら何もしないでENDCATCHに飛ぶ
                            state.JumpTo(func.JumpToEndCatch);
                    }
                    break;
                case FunctionCode.DOTRAIN:
                    {
                        switch (state.SystemState)
                        {
                            //case SystemStateCode.Train_Begin://BEGIN TRAINから。
                            case SystemStateCode.Train_CallEventTrain://@EVENTTRAINの呼び出し中。スキップ可能
                            case SystemStateCode.Train_CallShowStatus://@SHOW_STATUSの呼び出し中
                            //case SystemStateCode.Train_CallComAbleXX://@COM_ABLExxの呼び出し中。
                            case SystemStateCode.Train_CallShowUserCom://@SHOW_USERCOMの呼び出し中
                            //case SystemStateCode.Train_WaitInput://入力待ち状態。選択が実行可能ならEVENTCOMからCOMxx、そうでなければ@USERCOMにRESULTを渡す
                            //case SystemStateCode.Train_CallEventCom://@EVENTCOMの呼び出し中
                            //case SystemStateCode.Train_CallComXX://@COMxxの呼び出し中
                            //case SystemStateCode.Train_CallSourceCheck://@SOURCE_CHECKの呼び出し中
                            case SystemStateCode.Train_CallEventComEnd://@EVENTCOMENDの呼び出し中。スキップ可能。Train_CallEventTrainへ帰る。@USERCOMの呼び出し中もここ
                                break;
                            default:
                                throw new CodeEE("DOTRAIN命令をこの位置で実行することはできません");
                        }
                        coms.Clear();
                        isCTrain = false;
                        this.count = 0;

                        Int64 train = eEvaluator.GetInteger(((ExpressionArgument)func.Argument).Term);
                        if (train < 0)
                            throw new CodeEE("DOTRAIN命令に0未満の値が渡されました");
                        if (train >= vEvaluator.Constant.TrainName.Length)
                            throw new CodeEE("DOTRAIN命令にTRAINNAMEの配列数以上の値が渡されました");
                        doTrainSelectCom = train;
                        state.SystemState = SystemStateCode.Train_DoTrain;
                        return false;
                    }
                //case FunctionCode.LOOP:
                //    if (!sequential)
                //        break;
                //    expArg = (ExpressionArgument)func.Argument;
                //    if (eEvaluator.GetInteger(expArg.Term) == 0)//式が偽
                //        break;//そのまま次の処理へ
                //    state.JumpTo(func.JumpTo);
                //    break;
                default:
                    throw new ExeEE("未定義の関数です");
            }
            return true;
		}

		/// <summary>
		/// 呼び出し先関数への引数の代入。ローカル変数のScopeの都合上、JumpToやAddFunctionの前に呼ぶこと。
		/// </summary>
		/// <param name="called"></param>
		/// <param name="args"></param>
		private void assignArgs(FunctionLabelLine called, IOperandTerm[] args)
		{
            if (called.Arg.Length < args.Length)
                throw new CodeEE("引数の数が関数\"@" + called.LabelName + "\"に設定された数を超えています");
			if (called.Arg.Length == 0)
				return;
            string oldScope = null;
            if (state.CalledByFunction)
                oldScope = state.Scope;
			string newScope = called.LabelName;
			for (int i = 0; i < called.Arg.Length; i++)
			{
				IOperandTerm term = null;
				if (i < args.Length)
					term = args[i];
				FixedVariablePointer pt;
				if (term == null)
				{
					//VEvaluator.Scope = oldScope;
					//string str = eEvaluator.GetString(term);
					vEvaluator.Scope = newScope;
                    if (called.Arg[i].Identifier.Code == VariableCode.ARG || called.Arg[i].Identifier.Code == VariableCode.ARGS)
                    {
                        pt = eEvaluator.GetFixedVariable(called.Arg[i]);
                        if (called.Arg[i].GetOperandType() == typeof(string))
                            vEvaluator.SetValue(pt, eEvaluator.GetString(called.Def[i]));
                        else if (called.Arg[i].GetOperandType() == typeof(Int64))
                            vEvaluator.SetValue(pt, eEvaluator.GetInteger(called.Def[i]));
                        else
                            continue;
                    }
				}
				else if ((term.GetOperandType() == typeof(string)) && (called.Arg[i].GetOperandType() == typeof(string)))
				{
					vEvaluator.Scope = oldScope;
					string str = eEvaluator.GetString(term);
					vEvaluator.Scope = newScope;
					pt = eEvaluator.GetFixedVariable(called.Arg[i]);
					vEvaluator.SetValue(pt, str);
				}
				else if ((term.GetOperandType() == typeof(Int64)) && (called.Arg[i].GetOperandType() == typeof(Int64)))
				{
					vEvaluator.Scope = oldScope;
					Int64 l = eEvaluator.GetInteger(term);
					vEvaluator.Scope = newScope;
					pt = eEvaluator.GetFixedVariable(called.Arg[i]);
					vEvaluator.SetValue(pt, l);
				}
				else if ((term.GetOperandType() == typeof(Int64)) && (called.Arg[i].GetOperandType() == typeof(string)))
				{
					vEvaluator.Scope = oldScope;
					Int64 l = eEvaluator.GetInteger(term);
					vEvaluator.Scope = newScope;
					pt = eEvaluator.GetFixedVariable(called.Arg[i]);
					vEvaluator.SetValue(pt, l.ToString());
				}
				else if ((term.GetOperandType() == typeof(string)) && (called.Arg[i].GetOperandType() == typeof(Int64)))
					throw new CodeEE((i + 1).ToString() + "番目の引数を文字列型から整数型に変換できません");
				else
					throw new ExeEE("引数の形式が不明");
			}

		}

		//private List<string> splitArgStr(string str, FunctionCode bc)
		//{
		//    List<string> strArray = new List<string>();
		//    StringStream st = new StringStream(str);
		//    string tStr = null;
		//    int count = 0;
		//    while (!st.EOS)
		//    {
		//        TokenReader.SkipWhiteSpace(st);
		//        if (count > 0)
		//        {
		//            tStr = TokenReader.ReadIdentiferWithIndexAndExpression(st);
		//            strArray.Add(tStr);
		//            count++;
		//        }
		//        else
		//        {
		//            if (bc == FunctionCode.CALLFORM || bc == FunctionCode.TRYCALLFORM)
		//            {
		//                tStr = TokenReader.ReadRawStringWithFormString(st);
		//            }
		//            else
		//            {
		//                tStr = TokenReader.ReadRawString(st);
		//            }
		//            strArray.Add(tStr);
		//            count++;
		//        }
		//    }
		//    return strArray;
		//}


		List<ProcessState> prevStateList = new List<ProcessState>();
		private void saveCurrentState(bool single)
		{
			if (single && (prevStateList.Count > 0))
				throw new ExeEE("記憶している状態があるのに再度記憶しようとした");
			prevStateList.Add(state);
			state = state.Clone();
		}

		private void loadPrevState()
		{
			if (prevStateList.Count == 0)
				throw new ExeEE("記憶している状態がないのに呼び戻しされた");
			state = prevStateList[prevStateList.Count - 1];
			deletePrevState();
		}

		private void deletePrevState()
		{
			if (prevStateList.Count != 0)
				prevStateList.RemoveAt(prevStateList.Count - 1);
		}

		private void deleteAllPrevState()
		{
			prevStateList.Clear();
		}


		#endregion
	}
}
