using System;
using System.Collections.Generic;
using System.Text;
using MinorShift.Emuera.GameData.Function;

namespace MinorShift.Emuera.Sub
{


	/// <summary>
	/// FunctionCodeのラッパー
	/// BuiltInFunctionManagerを元にクラス化
	/// </summary>
	internal sealed class FunctionIdentifier
	{
		#region flag定義
		//いちいちFunctionFlag.FLOW_CONTROLとか書くのが面倒だったからenumではなくconst intで
		const int FLOW_CONTROL = 0x0001;
		const int EXTENDED = 0x0002;//Emuera拡張命令
		const int METHOD_SAFE = 0x0004;//#Function中で呼び出してよい命令。WAITなど入力を伴うもの、CALLなど関数呼び出しを伴うものは不可。
		const int DEBUG_FUNC = 0x0008;//-Debug引数付きで起動した場合にのみ実行される命令。

		const int PRINT_NEWLINE = 0x0010;//PRINT命令で出力後改行するもの
		const int PRINT_WAITINPUT = 0x0020;//PRINT命令で出力後入力待ちするもの
		const int PRINT_SINGLE = 0x0040;//PRINTSINGLE系
		const int ISPRINTDFUNC = 0x0080;//PRINTD系判定用
		const int ISPRINTKFUNC = 0x0100;//PRINTK系判定用
		#endregion

		#region static
		//元BuiltInFunctionManager部分
		readonly static Dictionary<string, FunctionIdentifier> funcDic = new Dictionary<string, FunctionIdentifier>();
		readonly static Dictionary<FunctionCode, string> funcMatch = new Dictionary<FunctionCode, string>();
		readonly static Dictionary<FunctionCode, FunctionCode> funcParent = new Dictionary<FunctionCode, FunctionCode>();
		private static void addFunction(FunctionCode code, FunctionArgType arg, int flag)
		{
			string key = code.ToString();
			if (Config.Instance.IgnoreCase)
				key = key.ToUpper();
			funcDic.Add(key, new FunctionIdentifier(code, arg, flag));
		}
		private static void addFunction(FunctionCode code, FunctionArgType arg)
		{
			string key = code.ToString();
			if (Config.Instance.IgnoreCase)
				key = key.ToUpper();
			funcDic.Add(key, new FunctionIdentifier(code, arg, 0));
		}
		private static void addMethod(string name)
		{
			string key = name;
			if (Config.Instance.IgnoreCase)
				key = key.ToUpper();
			funcDic.Add(key, new FunctionIdentifier(key, FunctionArgType.METHOD, METHOD_SAFE | EXTENDED));
		}
		static FunctionIdentifier()
		{

			setFunc = new FunctionIdentifier(FunctionCode.SET, FunctionArgType.SP_SET, METHOD_SAFE);//数値代入文
			setsFunc = new FunctionIdentifier(FunctionCode.SETS, FunctionArgType.SP_SETS, METHOD_SAFE);//文字列代入文

			addFunction(FunctionCode.PRINT, FunctionArgType.STR, METHOD_SAFE);//文字を表示する
			addFunction(FunctionCode.PRINTL, FunctionArgType.STR, METHOD_SAFE | PRINT_NEWLINE);//改行
			addFunction(FunctionCode.PRINTW, FunctionArgType.STR, PRINT_WAITINPUT);//入力待ち(実質改行)

			addFunction(FunctionCode.PRINTV, FunctionArgType.SP_PRINTV, METHOD_SAFE);//変数の内容
			addFunction(FunctionCode.PRINTVL, FunctionArgType.SP_PRINTV, METHOD_SAFE | PRINT_NEWLINE);
			addFunction(FunctionCode.PRINTVW, FunctionArgType.SP_PRINTV, PRINT_WAITINPUT);

			addFunction(FunctionCode.PRINTS, FunctionArgType.STR_EXPRESSION, METHOD_SAFE);//文字列変数の内容
			addFunction(FunctionCode.PRINTSL, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | PRINT_NEWLINE);//
			addFunction(FunctionCode.PRINTSW, FunctionArgType.STR_EXPRESSION, PRINT_WAITINPUT);//

			addFunction(FunctionCode.PRINTFORM, FunctionArgType.FORM_STR, METHOD_SAFE);//{数式}、%文字列変数%などの書式が使える。
			addFunction(FunctionCode.PRINTFORML, FunctionArgType.FORM_STR, METHOD_SAFE | PRINT_NEWLINE);
			addFunction(FunctionCode.PRINTFORMW, FunctionArgType.FORM_STR, PRINT_WAITINPUT);

			addFunction(FunctionCode.PRINTFORMS, FunctionArgType.STR_EXPRESSION, METHOD_SAFE);//文字列変数の内容を変換して表示。
			addFunction(FunctionCode.PRINTFORMSL, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | PRINT_NEWLINE);
			addFunction(FunctionCode.PRINTFORMSW, FunctionArgType.STR_EXPRESSION, PRINT_WAITINPUT);

			addFunction(FunctionCode.PRINTC, FunctionArgType.STR, METHOD_SAFE);//??

			addFunction(FunctionCode.CLEARLINE, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.REUSELASTLINE, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.WAIT, FunctionArgType.VOID);//改行待ち。
			addFunction(FunctionCode.INPUT, FunctionArgType.VOID);//整数入力。入力はRESULTへ。
			addFunction(FunctionCode.INPUTS, FunctionArgType.VOID);//文字列入力。入力はRESULTSへ。
			addFunction(FunctionCode.TINPUT, FunctionArgType.SP_BAR);
			addFunction(FunctionCode.TINPUTS, FunctionArgType.SP_TINPUTS);
			addFunction(FunctionCode.TWAIT, FunctionArgType.SP_SWAP, EXTENDED);
			addFunction(FunctionCode.ONEINPUT, FunctionArgType.VOID, EXTENDED);
			addFunction(FunctionCode.ONEINPUTS, FunctionArgType.VOID, EXTENDED);

			addFunction(FunctionCode.DRAWLINE, FunctionArgType.VOID, METHOD_SAFE);//画面の左端から右端まで----と線を引く。
			addFunction(FunctionCode.BAR, FunctionArgType.SP_BAR, METHOD_SAFE);//[*****....]のようなグラフを書く。BAR (変数) , (最大値), (長さ)
			addFunction(FunctionCode.BARL, FunctionArgType.SP_BAR, METHOD_SAFE);//改行付き。
			addFunction(FunctionCode.TIMES, FunctionArgType.SP_TIMES, METHOD_SAFE);//小数計算。TIMES (変数) , (小数値)という形で使う。

			addFunction(FunctionCode.PRINT_ABL, FunctionArgType.INT_EXPRESSION, METHOD_SAFE);//能力。引数は登録番号
			addFunction(FunctionCode.PRINT_TALENT, FunctionArgType.INT_EXPRESSION, METHOD_SAFE);//素質
			addFunction(FunctionCode.PRINT_MARK, FunctionArgType.INT_EXPRESSION, METHOD_SAFE);//刻印
			addFunction(FunctionCode.PRINT_EXP, FunctionArgType.INT_EXPRESSION, METHOD_SAFE);//経験
			addFunction(FunctionCode.PRINT_PALAM, FunctionArgType.INT_EXPRESSION, METHOD_SAFE);//パラメータ
			addFunction(FunctionCode.PRINT_ITEM, FunctionArgType.VOID, METHOD_SAFE);//所持アイテム
			addFunction(FunctionCode.PRINT_SHOPITEM, FunctionArgType.VOID, METHOD_SAFE);//ショップで売っているアイテム

			addFunction(FunctionCode.UPCHECK, FunctionArgType.VOID, METHOD_SAFE);//パラメータの変動
			addFunction(FunctionCode.CUPCHECK, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.ADDCHARA, FunctionArgType.INT_ANY, METHOD_SAFE);//(キャラ番号)のキャラクタを追加
			addFunction(FunctionCode.ADDSPCHARA, FunctionArgType.INT_ANY, METHOD_SAFE);//(キャラ番号)のSPキャラクタを追加（フラグ0を1にして作成）
			addFunction(FunctionCode.ADDDEFCHARA, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.ADDVOIDCHARA, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);//変数に何の設定のないキャラを作成
			addFunction(FunctionCode.DELCHARA, FunctionArgType.INT_ANY, METHOD_SAFE);//(キャラ登録番号)のキャラクタを削除。

			addFunction(FunctionCode.PUTFORM, FunctionArgType.FORM_STR, METHOD_SAFE);//@SAVEINFO関数でのみ使用可能。PRINTFORMと同様の書式でセーブデータに概要をつける。
			addFunction(FunctionCode.QUIT, FunctionArgType.VOID);//ゲームを終了
			addFunction(FunctionCode.OUTPUTLOG, FunctionArgType.VOID);

			addFunction(FunctionCode.BEGIN, FunctionArgType.STR, FLOW_CONTROL);//システム関数の実行。実行するとCALLの呼び出し元などを忘れてしまう。

			addFunction(FunctionCode.SAVEGAME, FunctionArgType.VOID, FLOW_CONTROL);//セーブ画面を呼ぶ。ショップのみ。
			addFunction(FunctionCode.LOADGAME, FunctionArgType.VOID, FLOW_CONTROL);//

			addFunction(FunctionCode.SIF, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | FLOW_CONTROL);//一行のみIF
			addFunction(FunctionCode.IF, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | FLOW_CONTROL);
			addFunction(FunctionCode.ELSE, FunctionArgType.VOID, METHOD_SAFE | FLOW_CONTROL);
			addFunction(FunctionCode.ELSEIF, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | FLOW_CONTROL);
			addFunction(FunctionCode.ENDIF, FunctionArgType.VOID, METHOD_SAFE | FLOW_CONTROL);

			addFunction(FunctionCode.REPEAT, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | FLOW_CONTROL);//RENDまで繰り返し。繰り返した回数がCOUNTへ。ネスト不可。
			addFunction(FunctionCode.REND, FunctionArgType.VOID, METHOD_SAFE | FLOW_CONTROL);
			addFunction(FunctionCode.CONTINUE, FunctionArgType.VOID, METHOD_SAFE | FLOW_CONTROL);//REPEATに戻る
			addFunction(FunctionCode.BREAK, FunctionArgType.VOID, METHOD_SAFE | FLOW_CONTROL);//RENDの次の行まで

			addFunction(FunctionCode.GOTO, FunctionArgType.STR, METHOD_SAFE | FLOW_CONTROL);//$ラベルへジャンプ

			addFunction(FunctionCode.JUMP, FunctionArgType.SP_CALL, FLOW_CONTROL);//関数に移動
			addFunction(FunctionCode.CALL, FunctionArgType.SP_CALL, FLOW_CONTROL);//関数に移動。移動元を記憶し、RETURNで帰る。
			addFunction(FunctionCode.RETURN, FunctionArgType.INT_ANY, FLOW_CONTROL);//関数の終了。RESULTに整数を格納可能。省略した場合、０。(次の@～～がRETURNと見なされる。)  
			addFunction(FunctionCode.RETURNFORM, FunctionArgType.FORM_STR_ANY, EXTENDED | FLOW_CONTROL);//関数の終了。RESULTに整数を格納可能。省略した場合、０。(次の@～～がRETURNと見なされる。)  
			addFunction(FunctionCode.RETURNF, FunctionArgType.EXPRESSION, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.RESTART, FunctionArgType.VOID, FLOW_CONTROL);//関数の再開。関数の最初に戻る。


			addFunction(FunctionCode.STRLEN, FunctionArgType.STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.STRLENFORM, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.STRLENU, FunctionArgType.STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.STRLENFORMU, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.PRINTLC, FunctionArgType.STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.PRINTFORMC, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.PRINTFORMLC, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.SWAPCHARA, FunctionArgType.SP_SWAP, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.VARSIZE, FunctionArgType.SP_VAR, METHOD_SAFE | EXTENDED);//動作が違うのでMETHOD化できない
			addFunction(FunctionCode.SPLIT, FunctionArgType.SP_SPLIT, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.PRINTSINGLE, FunctionArgType.STR, METHOD_SAFE | EXTENDED | PRINT_SINGLE);
			addFunction(FunctionCode.PRINTSINGLEV, FunctionArgType.SP_PRINTV, METHOD_SAFE | EXTENDED | PRINT_SINGLE);
			addFunction(FunctionCode.PRINTSINGLES, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_SINGLE);
			addFunction(FunctionCode.PRINTSINGLEFORM, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | PRINT_SINGLE);
			addFunction(FunctionCode.PRINTSINGLEFORMS, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_SINGLE);

			addFunction(FunctionCode.PRINTBUTTON, FunctionArgType.SP_BUTTON, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.PRINTBUTTONC, FunctionArgType.SP_BUTTON, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.PRINTBUTTONLC, FunctionArgType.SP_BUTTON, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.PRINTPLAIN, FunctionArgType.STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.PRINTPLAINFORM, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.SAVEDATA, FunctionArgType.SP_SAVEDATA, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.LOADDATA, FunctionArgType.INT_EXPRESSION, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.DELDATA, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.GETTIME, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);//2つに代入する必要があるのでMETHOD化できない

			addFunction(FunctionCode.TRYJUMP, FunctionArgType.SP_CALL, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCALL, FunctionArgType.SP_CALL, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYGOTO, FunctionArgType.STR, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.JUMPFORM, FunctionArgType.SP_CALLFORM, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.CALLFORM, FunctionArgType.SP_CALLFORM, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.GOTOFORM, FunctionArgType.FORM_STR, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYJUMPFORM, FunctionArgType.SP_CALLFORM, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCALLFORM, FunctionArgType.SP_CALLFORM, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYGOTOFORM, FunctionArgType.FORM_STR, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.CALLTRAIN, FunctionArgType.INT_EXPRESSION, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.CATCH, FunctionArgType.VOID, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.ENDCATCH, FunctionArgType.VOID, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCJUMP, FunctionArgType.SP_CALL, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCCALL, FunctionArgType.SP_CALL, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCGOTO, FunctionArgType.STR, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCJUMPFORM, FunctionArgType.SP_CALLFORM, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCCALLFORM, FunctionArgType.SP_CALLFORM, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCGOTOFORM, FunctionArgType.FORM_STR, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYCALLLIST, FunctionArgType.VOID, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYJUMPLIST, FunctionArgType.VOID, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.TRYGOTOLIST, FunctionArgType.VOID, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.FUNC, FunctionArgType.SP_CALLFORM, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.ENDFUNC, FunctionArgType.VOID, EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.CALLF, FunctionArgType.SP_CALL, EXTENDED | METHOD_SAFE);
			addFunction(FunctionCode.CALLFORMF, FunctionArgType.SP_CALLFORM, EXTENDED | METHOD_SAFE);

			addFunction(FunctionCode.SETCOLOR, FunctionArgType.SP_COLOR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.RESETCOLOR, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.FONTBOLD, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.FONTITALIC, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.FONTREGULAR, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.SORTCHARA, FunctionArgType.SP_SORTCHARA, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.FONTSTYLE, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.ALIGNMENT, FunctionArgType.STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.CUSTOMDRAWLINE, FunctionArgType.STR, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.SETFONT, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.FOR, FunctionArgType.SP_FOR_NEXT, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.NEXT, FunctionArgType.VOID, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.WHILE, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.WEND, FunctionArgType.VOID, METHOD_SAFE | EXTENDED | FLOW_CONTROL);

			addFunction(FunctionCode.POWER, FunctionArgType.SP_POWER, METHOD_SAFE | EXTENDED);//引数が違うのでMETHOD化できない。
			addFunction(FunctionCode.SAVEGLOBAL, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.LOADGLOBAL, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.SWAP, FunctionArgType.SP_SWAPVAR, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.RESETDATA, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.RESETGLOBAL, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.RANDOMIZE, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.DUMPRAND, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.INITRAND, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.REDRAW, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.DOTRAIN, FunctionArgType.INT_EXPRESSION, EXTENDED | FLOW_CONTROL);

			addFunction(FunctionCode.SELECTCASE, FunctionArgType.EXPRESSION, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.CASE, FunctionArgType.CASE, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.CASEELSE, FunctionArgType.VOID, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.ENDSELECT, FunctionArgType.VOID, METHOD_SAFE | EXTENDED | FLOW_CONTROL);

			addFunction(FunctionCode.DO, FunctionArgType.VOID, METHOD_SAFE | EXTENDED | FLOW_CONTROL);
			addFunction(FunctionCode.LOOP, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED | FLOW_CONTROL);

			addFunction(FunctionCode.PRINTDATA, FunctionArgType.VAR_INT, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.PRINTDATAL, FunctionArgType.VAR_INT, METHOD_SAFE | EXTENDED | PRINT_NEWLINE);
			addFunction(FunctionCode.PRINTDATAW, FunctionArgType.VAR_INT, EXTENDED | PRINT_WAITINPUT);
			addFunction(FunctionCode.DATA, FunctionArgType.STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.DATAFORM, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.ENDDATA, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.DATALIST, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.ENDLIST, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.PRINTCPERLINE, FunctionArgType.SP_GETINT, METHOD_SAFE | EXTENDED);//よく考えたら引数の仕様違うや

			addFunction(FunctionCode.SETBIT, FunctionArgType.BIT_ARG, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.CLEARBIT, FunctionArgType.BIT_ARG, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.INVERTBIT, FunctionArgType.BIT_ARG, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.DELALLCHARA, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.PICKUPCHARA, FunctionArgType.INT_ANY, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.VARSET, FunctionArgType.SP_VAR_SET, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.CVARSET, FunctionArgType.SP_CVAR_SET, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.RESET_STAIN, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.SAVENOS, FunctionArgType.SP_GETINT, METHOD_SAFE | EXTENDED);//引数の仕様が違うので(ry

			addFunction(FunctionCode.FORCEKANA, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.SKIPDISP, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.NOSKIP, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.ENDNOSKIP, FunctionArgType.VOID, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.ARRAYSHIFT, FunctionArgType.SP_SHIFT_ARRAY, METHOD_SAFE | EXTENDED);
			addFunction(FunctionCode.ARRAYREMOVE, FunctionArgType.SP_CONTROL_ARRAY, METHOD_SAFE | EXTENDED);

			addFunction(FunctionCode.ENCODETOUNI, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED);


			addFunction(FunctionCode.DEBUGPRINT, FunctionArgType.STR, METHOD_SAFE | EXTENDED | DEBUG_FUNC);
			addFunction(FunctionCode.DEBUGPRINTL, FunctionArgType.STR, METHOD_SAFE | EXTENDED | DEBUG_FUNC | PRINT_NEWLINE);
			addFunction(FunctionCode.DEBUGPRINTFORM, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | DEBUG_FUNC);
			addFunction(FunctionCode.DEBUGPRINTFORML, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | DEBUG_FUNC | PRINT_NEWLINE);
			addFunction(FunctionCode.ASSERT, FunctionArgType.INT_EXPRESSION, METHOD_SAFE | EXTENDED | DEBUG_FUNC);

			addFunction(FunctionCode.PRINTK, FunctionArgType.STR, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTKL, FunctionArgType.STR, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTKW, FunctionArgType.STR, EXTENDED | PRINT_WAITINPUT | ISPRINTKFUNC);

			addFunction(FunctionCode.PRINTVK, FunctionArgType.SP_PRINTV, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);//変数の内容
			addFunction(FunctionCode.PRINTVKL, FunctionArgType.SP_PRINTV, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTVKW, FunctionArgType.SP_PRINTV, EXTENDED | PRINT_WAITINPUT | ISPRINTKFUNC);

			addFunction(FunctionCode.PRINTSK, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);//文字列変数の内容
			addFunction(FunctionCode.PRINTSKL, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTSKW, FunctionArgType.STR_EXPRESSION, EXTENDED | PRINT_WAITINPUT | ISPRINTKFUNC);

			addFunction(FunctionCode.PRINTFORMK, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);//{数式}、%文字列変数%などの書式が使える。
			addFunction(FunctionCode.PRINTFORMKL, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTFORMKW, FunctionArgType.FORM_STR, EXTENDED | PRINT_WAITINPUT | ISPRINTKFUNC);

			addFunction(FunctionCode.PRINTFORMSK, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);//文字列変数の内容を変換して表示。
			addFunction(FunctionCode.PRINTFORMSKL, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTFORMSKW, FunctionArgType.STR_EXPRESSION, EXTENDED | PRINT_WAITINPUT | ISPRINTKFUNC);

			addFunction(FunctionCode.PRINTCK, FunctionArgType.STR, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);//??
			addFunction(FunctionCode.PRINTLCK, FunctionArgType.STR, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTFORMCK, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTFORMLCK, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);

			addFunction(FunctionCode.PRINTSINGLEK, FunctionArgType.STR, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTSINGLEVK, FunctionArgType.SP_PRINTV, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTSINGLESK, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTSINGLEFORMK, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTSINGLEFORMSK, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTKFUNC);

			addFunction(FunctionCode.PRINTDATAK, FunctionArgType.VAR_INT, METHOD_SAFE | EXTENDED | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTDATAKL, FunctionArgType.VAR_INT, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTKFUNC);
			addFunction(FunctionCode.PRINTDATAKW, FunctionArgType.VAR_INT, EXTENDED | PRINT_WAITINPUT | ISPRINTKFUNC);

			addFunction(FunctionCode.PRINTD, FunctionArgType.STR, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);//文字を表示する
			addFunction(FunctionCode.PRINTDL, FunctionArgType.STR, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTDFUNC);//改行
			addFunction(FunctionCode.PRINTDW, FunctionArgType.STR, EXTENDED | PRINT_WAITINPUT | ISPRINTDFUNC);//入力待ち(実質改行)

			addFunction(FunctionCode.PRINTVD, FunctionArgType.SP_PRINTV, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);//変数の内容
			addFunction(FunctionCode.PRINTVDL, FunctionArgType.SP_PRINTV, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTVDW, FunctionArgType.SP_PRINTV, EXTENDED | PRINT_WAITINPUT | ISPRINTDFUNC);

			addFunction(FunctionCode.PRINTSD, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);//文字列変数の内容
			addFunction(FunctionCode.PRINTSDL, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTSDW, FunctionArgType.STR_EXPRESSION, EXTENDED | PRINT_WAITINPUT | ISPRINTDFUNC);

			addFunction(FunctionCode.PRINTFORMD, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);//{数式}、%文字列変数%などの書式が使える。
			addFunction(FunctionCode.PRINTFORMDL, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTFORMDW, FunctionArgType.FORM_STR, EXTENDED | PRINT_WAITINPUT | ISPRINTDFUNC);

			addFunction(FunctionCode.PRINTFORMSD, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);//文字列変数の内容を変換して表示。
			addFunction(FunctionCode.PRINTFORMSDL, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTFORMSDW, FunctionArgType.STR_EXPRESSION, EXTENDED | PRINT_WAITINPUT | ISPRINTDFUNC);

			addFunction(FunctionCode.PRINTCD, FunctionArgType.STR, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);//??
			addFunction(FunctionCode.PRINTLCD, FunctionArgType.STR, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTFORMCD, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTFORMLCD, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);

			addFunction(FunctionCode.PRINTSINGLED, FunctionArgType.STR, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTSINGLEVD, FunctionArgType.SP_PRINTV, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTSINGLESD, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTSINGLEFORMD, FunctionArgType.FORM_STR, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTSINGLEFORMSD, FunctionArgType.STR_EXPRESSION, METHOD_SAFE | EXTENDED | PRINT_SINGLE | ISPRINTDFUNC);

			addFunction(FunctionCode.PRINTDATAD, FunctionArgType.VAR_INT, METHOD_SAFE | EXTENDED | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTDATADL, FunctionArgType.VAR_INT, METHOD_SAFE | EXTENDED | PRINT_NEWLINE | ISPRINTDFUNC);
			addFunction(FunctionCode.PRINTDATADW, FunctionArgType.VAR_INT, EXTENDED | PRINT_WAITINPUT | ISPRINTDFUNC);

			foreach (string methodName in FunctionMethodCreator.GetMethodNameList())
			{
				if (!funcDic.ContainsKey(methodName))
				{
					addMethod(methodName);
				}
			}
			funcMatch[FunctionCode.IF] = "ENDIF";
			funcMatch[FunctionCode.SELECTCASE] = "ENDSELECT";
			funcMatch[FunctionCode.REPEAT] = "REND";
			funcMatch[FunctionCode.FOR] = "NEXT";
			funcMatch[FunctionCode.WHILE] = "WEND";
			funcMatch[FunctionCode.TRYCGOTO] = "CATCH";
			funcMatch[FunctionCode.TRYCJUMP] = "CATCH";
			funcMatch[FunctionCode.TRYCCALL] = "CATCH";
			funcMatch[FunctionCode.TRYCGOTOFORM] = "CATCH";
			funcMatch[FunctionCode.TRYCJUMPFORM] = "CATCH";
			funcMatch[FunctionCode.TRYCCALLFORM] = "CATCH";
			funcMatch[FunctionCode.DO] = "LOOP";
			funcMatch[FunctionCode.PRINTDATA] = "ENDDATA";
			funcMatch[FunctionCode.PRINTDATAL] = "ENDDATA";
			funcMatch[FunctionCode.PRINTDATAW] = "ENDDATA";
			funcMatch[FunctionCode.NOSKIP] = "ENDNOSKIP";
			funcParent[FunctionCode.REND] = FunctionCode.REPEAT;
			funcParent[FunctionCode.NEXT] = FunctionCode.FOR;
			funcParent[FunctionCode.WEND] = FunctionCode.WHILE;
			funcParent[FunctionCode.LOOP] = FunctionCode.DO;
		}

		public static FunctionIdentifier ToBuiltInFunction(string str)
		{
			string key = str;
			if (string.IsNullOrEmpty(key))
				return null;
			if (Config.Instance.IgnoreCase)
				key = key.ToUpper();
			if (!funcDic.ContainsKey(key))
				return null;
			return funcDic[key];
		}

		private static FunctionIdentifier setFunc;
		private static FunctionIdentifier setsFunc;

		public static FunctionIdentifier GetSETFunction()
		{
			return setFunc;
		}
		public static FunctionIdentifier GetSETSFunction()
		{
			return setsFunc;
		}

		internal static string getMatchFunction(FunctionCode func)
		{
			if (funcMatch.ContainsKey(func))
				return funcMatch[func];
			return null;
		}


		internal static FunctionCode getParentFunc(FunctionCode func)
		{
			//1755 どうもenum.ToString()が遅いようなので先に逆引き辞書を作ることに
			if (funcParent.ContainsKey(func))
				return funcParent[func];
			return FunctionCode.__NULL__;
			//if (funcMatch.ContainsValue(func.ToString()))
			//{
			//    foreach (FunctionCode pFunc in funcMatch.Keys)
			//    {
			//        if (funcMatch[pFunc] == func.ToString())
			//            return pFunc;
			//    }
			//}
			//return FunctionCode.__NULL__;
		}
		#endregion

		private FunctionIdentifier(FunctionCode code, FunctionArgType arg, int flag)
		{ this.code = code; this.arg = arg; this.flag = flag; this.methodName = null;}
		private FunctionIdentifier(string methodName, FunctionArgType arg, int flag)
		{ this.code = FunctionCode.__NULL__; this.arg = arg; this.flag = flag; this.methodName = methodName;}

		private FunctionCode code;
		public FunctionCode Code { get { return code; } }
		private FunctionArgType arg;
		public FunctionArgType ArgType { get { return arg; } }
		private int flag;
		private string methodName = null;
		public string MethodName { get { return methodName; } }

		internal bool IsFlowContorol()
		{
			return ((flag & FLOW_CONTROL) == FLOW_CONTROL);
		}

		internal bool IsPrintDFunction()
		{
			return ((flag & ISPRINTDFUNC) == ISPRINTDFUNC);
		}

		internal bool IsPrintKFunction()
		{
			return ((flag & ISPRINTDFUNC) == ISPRINTKFUNC);
		}

		internal bool IsNewLine()
		{
			return ((flag & PRINT_NEWLINE) == PRINT_NEWLINE);
		}

		internal bool IsWaitInput()
		{
			return ((flag & PRINT_WAITINPUT) == PRINT_WAITINPUT);
		}
		internal bool IsPrintSingle()
		{
			return ((flag & PRINT_SINGLE) == PRINT_SINGLE);
		}

		internal bool IsMethodSafe()
		{
			return ((flag & METHOD_SAFE) == METHOD_SAFE);
		}
		internal bool IsDebug()
		{
			return ((flag & DEBUG_FUNC) == DEBUG_FUNC);
		}

		internal bool IsMethod()
		{
			return methodName != null;
		}

		public override string ToString()
		{
			if (methodName != null)
				return methodName;
			return code.ToString();
		}

	}
}